#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define VPCALC_SECTION	"CPU"
#define VPCALC_SUBSECT	"Z80"

#include "../../config.h"
#include "../../include/types.h"
#include "../../include/debug.h"

#include "../../include/timing.h"

#include "../../include/mem.h"
#include "../../include/gfx.h"
#include "../../include/dev.h"

#include "../../include/core.h"

#include "z80.h"
#include "flags.h"

#define LD_R(_core)	((uint8)((_core->r << 7) | (_core->r >> 1)))
#define ST_R(val)	((uint8)((val << 1) | (val >> 7)))

//#define DEBUG 1
//#define DEBUG_USE 1

static inline void bad_opcode(_core_o *_core, unsigned int prefix)
{
	_err_report("Unsupported Opcode: %X%02X", prefix, _core->current_opc);
}

#ifdef DEBUG_USE
static int _used_main[256];
static int _used_extended[256];
static int _used_indexed[256];
static int _used_bit[256];
#endif

core_o core_new(mem_o mem, dev_o dev, timing_o *rate)
{
	_core_o *new;

#ifdef DEBUG_USE
        memset(_used_main, 0, sizeof(_used_main));
        memset(_used_extended, 0, sizeof(_used_extended));
        memset(_used_indexed, 0, sizeof(_used_indexed));
        memset(_used_bit, 0, sizeof(_used_bit));
#endif

	if (mem_bus_size(mem) != 8) {
		fprintf(stderr, "[Z80/INIT] Unable to use a memory unit with a bus size that is not 1!\n");
		return NULL;
	}

	new = (_core_o *) malloc(sizeof(_core_o));

	if (new != NULL) {
		new->bc.i16 = 0xFFFF;
		new->de.i16 = 0xFFFF;
		new->hl.i16 = 0xFFFF;
		new->af.i16 = 0xFFFF;

		new->sp = 0xFFFF;
		new->pc = 0;

		new->ix.i16 = 0xFFFF;
		new->iy.i16 = 0xFFFF;

		new->i = 0xFF;
		new->r = 0xFF;

		new->im = 0;
		new->iff1 = 0;
		new->iff2 = 0;

		new->halted = 0;
		new->active = 0;
		new->interrupted = 0;

		new->addr = 0xFFFF;
		new->ii = &(new->hl);

		new->current_opc = 0;

		new->raw_avail = mem_raw(mem, (void **)&new->rd,
			(void **)&new->wr,(void **)&new->ex);

		new->timings = rate;
		new->low_power = 0;

		new->cycles = new->timings->cpint;
		new->terminate = 0;

		new->mem = mem;
		new->dev = dev;

		dev_map_sleep(dev, &(new->low_power));
	}

	return (core_o)new;
}

void core_del(core_o core)
{
	_core_o *_core = (_core_o *)core;
                                                                                
	free(_core);
}

//returns the byte at the address specified
INTERNAL uint8 _ld_b(_core_o *_core, uint16 addr)
{
	_cycle(_core, 3);

	if (!_core->raw_avail)
		return mem_read(_core->mem, addr);

	return _core->rd[addr & 0xffff];
}

//stores the byte at the address specified
INTERNAL void _st_b(_core_o *_core, uint16 addr, uint8 data)
{
	_cycle(_core, 3);

	if (!_core->raw_avail)
		mem_write(_core->mem, addr, data);
	else
		_core->wr[addr & 0xffff] = data;
}

//returns the word at the address specified
INTERNAL uint16 _ld_w(_core_o *_core, uint16 addr)
{
	reg_pair tmp;

	tmp.i8.l = _ld_b(_core, addr);
	tmp.i8.h = _ld_b(_core, addr+1);

	return tmp.i16;
}

//stores the word at the address specified
INTERNAL void _st_w(_core_o *_core, uint16 addr, uint16 data)
{
	reg_pair tmp;

	tmp.i16 = data;

	_st_b(_core, addr, tmp.i8.l);
	_st_b(_core, addr+1, tmp.i8.h);
}

//returns the byte immediate at PC, incrementing it
INTERNAL uint8 _imm_b(_core_o *_core)
{
	return _ld_b(_core, _core->pc++);
}

//returns the opcode at PC, incrementing it
INTERNAL void _fetch(_core_o *_core)
{
	_cycle(_core, 4);

	if (!_core->raw_avail)
		_core->current_opc = mem_fetch(_core->mem, _core->pc++);
	else
		_core->current_opc = _core->ex[_core->pc++ & 0xffff];
}

//returns the sign-extended byte immediate at PC, incrementing it
INTERNAL uint16 _imm_sb(_core_o *_core)
{
	sint16 ret = (sint8)_imm_b(_core);

	return (uint16)ret;
}

//returns the word immediate at PC, incrementing it
INTERNAL uint16 _imm_w(_core_o *_core)
{
	uint16 wtmp;

	wtmp = _ld_w(_core, _core->pc);
	_core->pc += 2;

	return wtmp;
}

//returns a 16-bit bit addition
INTERNAL uint16 _add_i16(_core_o *_core, uint16 a, uint16 b)
{
	_core->addr = a + b;
	_cycle(_core, 5);

	return _core->addr;
}

//returns the calculated address of ii+d
INTERNAL uint16 _addr_ii(_core_o *_core)
{
	return _add_i16(_core, _core->ii->i16, _imm_sb(_core));
}

//returns VFLAG if the parity of the operand is even
INTERNAL uint8 _parity(uint8 val)
{
	val = val ^ (val >> 1);
	val = val ^ (val >> 2);
	val = val ^ (val >> 4);

	return (~val) & 0x01;
}

/*	nop
 *	.190
 */
static void m00(_core_o *_core)
{
}

/*	undocumented-nop
 *	.190
 */
static void x00(_core_o *_core)
{
}

/*	ex af, af'
 *	.141
 */
static void m08(_core_o *_core)
{
	uint16 wtmp = _core->af.i16;
	_core->af.i16 = _core->af_;
	_core->af_ = wtmp;
}

INTERNAL void _jr(_core_o *_core, uint16 offset)
{
	_core->pc = _add_i16(_core, _core->pc, offset);
}

/*	djnz e
 *	.271
 */
static void m10(_core_o *_core)
{
	uint16 offset;

	offset = _imm_sb(_core);

	_cycle(_core, 1);
	if (--_core->bc.i8.h == 0)
		return;

	_jr(_core, offset);
}

/*	jr e
 *	.259
 */
static void m18(_core_o *_core)
{
	uint16 offset = _imm_sb(_core);

	return _jr(_core, offset);
}

INTERNAL void _jr_f(_core_o *_core, uint8 flag)
{
	uint16 offset = _imm_sb(_core);

	if (_core->af.i8.l & flag)
		_jr(_core, offset);
}

INTERNAL void _jr_nf(_core_o *_core, uint8 flag)
{
	uint16 offset = _imm_sb(_core);

	if (!(_core->af.i8.l & flag))
		_jr(_core, offset);
}

/*	jr nc, e
 *	.262
 */
static void m30(_core_o *_core)
{
	_jr_nf(_core, CFLAG);
}

/*	jr cf, e
 *	.260
 */
static void m38(_core_o *_core)
{
	_jr_f(_core, CFLAG);
}

/*	jr nz, e
 *	.266
 */
static void m20(_core_o *_core)
{
	_jr_nf(_core, ZFLAG);
}

/*	jr z, e
 *	.266
 */
static void m28(_core_o *_core)
{
	_jr_f(_core, ZFLAG);
}

/*	ld bc, nnnn
 *	.120
 */
static void m01(_core_o *_core)
{
	_core->bc.i16 = _imm_w(_core);
}

/*	ld de, nn
 *	.120
 */
static void m11(_core_o *_core)
{
	_core->de.i16 = _imm_w(_core);
}

/*	ld ii, nn
 *	.120
 */
static void m21(_core_o *_core)
{
	_core->ii->i16 = _imm_w(_core);
}

/*	ld sp, nn
 *	.120
 */
static void m31(_core_o *_core)
{
	_core->sp = _imm_w(_core);
}

/*	ld (bc), a
 *	.113
 */
static void m02(_core_o *_core)
{
	_st_b(_core, _core->bc.i16, _core->af.i8.h);
}

/*	ld a, (bc)
 *	.110
 */
static void m0A(_core_o *_core)
{
	_core->af.i8.h = _ld_b(_core, _core->bc.i16);
}

/*	ld (de), a
 *	.114
 */
static void m12(_core_o *_core)
{
	_st_b(_core, _core->de.i16, _core->af.i8.h);
}

/*	ld a, (de)
 *	.111
 */
static void m1A(_core_o *_core)
{
	_core->af.i8.h = _ld_b(_core, _core->de.i16);
}

/*	ld (nn), ii
 *	.127
 */
static void m22(_core_o *_core)
{
	uint16 addr = _imm_w(_core);

	_st_w(_core, addr, _core->ii->i16);
}

/*	ld ii, (nn)
 *	.123
 */
static void m2A(_core_o *_core)
{
	uint16 addr = _imm_w(_core);

	_core->ii->i16 = _ld_w(_core, addr);
}

/*	ld (nn), a
 *	.115
 */
static void m32(_core_o *_core)
{
	uint16 addr = _imm_w(_core);

	_st_b(_core, addr, _core->af.i8.h);
}

/*	ld a, (nn)
 *	.112
 */
static void m3A(_core_o *_core)
{
	uint16 addr = _imm_w(_core);

	_core->af.i8.h = _ld_b(_core, addr);
}

/*	inc bc
 *	.202
 */
static void m03(_core_o *_core)
{
	_core->bc.i16++;

	_cycle(_core, 2);
}

/*	inc de
 *	.202
 */
static void m13(_core_o *_core)
{
	_core->de.i16++;

	_cycle(_core, 2);
}

/*	inc hl
 *	.202
 */
static void m23(_core_o *_core)
{
	_core->ii->i16++;

	_cycle(_core, 2);
}

/*	inc sp
 *	.202
 */
static void m33(_core_o *_core)
{
	_core->sp++;

	_cycle(_core, 2);
}

/*	dec bc
 *	.205
 */
static void m0B(_core_o *_core)
{
	_core->bc.i16--;

	_cycle(_core, 2);
}

/*	dec de
 *	.205
 */
static void m1B(_core_o *_core)
{
	_core->de.i16--;

	_cycle(_core, 2);
}

/*	dec hl
 *	.205
 */
static void m2B(_core_o *_core)
{
	_core->ii->i16--;

	_cycle(_core, 2);
}

/*	dec sp
 *	.205
 */
static void m3B(_core_o *_core)
{
	_core->sp--;

	_cycle(_core, 2);
}

INTERNAL void _f_inc(_core_o *_core, uint8 val)
{
	uint8 f;

	f = (val & (SFLAG | BITS35));

	if (!val)
		f |= ZFLAG;

	if ((val & 0x0f) == 0x00)
		f |= HFLAG;

	if (val == 0x80)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | f;
}

/*	inc b
 *	.178
 */
static void m04(_core_o *_core)
{
	_f_inc(_core, ++_core->bc.i8.h);
}
/*	inc c
 *	.178
 */
static void m0C(_core_o *_core)
{
	_f_inc(_core, ++_core->bc.i8.l);
}

/*	inc d
 *	.178
 */
static void m14(_core_o *_core)
{
	_f_inc(_core, ++_core->de.i8.h);
}
/*	inc e
 *	.178
 */
static void m1C(_core_o *_core)
{
	_f_inc(_core, ++_core->de.i8.l);
}

/*	inc h
 *	.178
 */
static void m24(_core_o *_core)
{
	_f_inc(_core, ++_core->ii->i8.h);
}
/*	inc l
 *	.178
 */
static void m2C(_core_o *_core)
{
	_f_inc(_core, ++_core->ii->i8.l);
}

/*	inc (ii+d)
 *	.180
 */
static void i34(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);

	btmp = _ld_b(_core, addr);
	_f_inc(_core, ++btmp);
	_cycle(_core, 1);
	_st_b(_core, addr, btmp);
}

/*	inc (hl)
 *	.179
 */
static void m34(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_f_inc(_core, ++btmp);
	_cycle(_core, 1);
	_st_b(_core, _core->hl.i16, btmp);
}
/*	inc a
 *	.178
 */
static void m3C(_core_o *_core)
{
	_f_inc(_core, ++_core->af.i8.h);
}

INTERNAL void _f_dec(_core_o *_core, uint8 val)
{
	uint8 f;

	f = (val & (SFLAG | BITS35)) | NFLAG;

	if (!val)
		f |= ZFLAG;

	if ((val & 0x0f) == 0x0f)
		f |= HFLAG;

	if (val == 0x7F)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | f;
}

/*	dec b
 *	.182
 */
static void m05(_core_o *_core)
{
	_f_dec(_core, --_core->bc.i8.h);
}
/*	dec c
 *	.182
 */
static void m0D(_core_o *_core)
{
	_f_dec(_core, --_core->bc.i8.l);
}

/*	dec d
 *	.182
 */
static void m15(_core_o *_core)
{
	_f_dec(_core, --_core->de.i8.h);
}
/*	dec e
 *	.182
 */
static void m1D(_core_o *_core)
{
	_f_dec(_core, --_core->de.i8.l);
}

/*	dec h
 *	.182
 */
static void m25(_core_o *_core)
{
	_f_dec(_core, --_core->ii->i8.h);
}
/*	dec l
 *	.182
 */
static void m2D(_core_o *_core)
{
	_f_dec(_core, --_core->ii->i8.l);
}

/*	dec (ii+d)
 *	.182
 */
static void i35(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);

	btmp = _ld_b(_core, addr);
	_f_dec(_core, --btmp);
	_cycle(_core, 1);
	_st_b(_core, addr, btmp);
}

/*	dec (hl)
 *	.182
 */
static void m35(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_f_dec(_core, --btmp);
	_cycle(_core, 1);
	_st_b(_core, _core->hl.i16, btmp);
}
/*	dec a
 *	.182
 */
static void m3D(_core_o *_core)
{
	_f_dec(_core, --_core->af.i8.h);
}

/*	ld b, n
 *	.100
 */
static void m06(_core_o *_core)
{
	_core->bc.i8.h = _imm_b(_core);
}
/*	ld c, n
 *	.100
 */
static void m0E(_core_o *_core)
{
	_core->bc.i8.l = _imm_b(_core);
}

/*	ld d, n
 *	.100
 */
static void m16(_core_o *_core)
{
	_core->de.i8.h = _imm_b(_core);
}
/*	ld e, n
 *	.100
 */
static void m1E(_core_o *_core)
{
	_core->de.i8.l = _imm_b(_core);
}

/*	ld h, n
 *	.100
 */
static void m26(_core_o *_core)
{
	_core->ii->i8.h = _imm_b(_core);
}
/*	ld l, n
 *	.100
 */
static void m2E(_core_o *_core)
{
	_core->ii->i8.l = _imm_b(_core);
}

/*	ld (ii+d), n
 *	.108
 */
static void i36(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _imm_b(_core));
}

/*	ld (hl), n
 *	.100
 */
static void m36(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _imm_b(_core));
}
/*	ld a, n
 *	.100
 */
static void m3E(_core_o *_core)
{
	_core->af.i8.h = _imm_b(_core);
}

/*	rlca
 *	.208
 */
static void m07(_core_o *_core)
{
	uint8 btmp = _core->af.i8.h;

	_core->af.i8.h = (btmp << 1) | (btmp >> 7);

	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | VFLAG)) |
		(_core->af.i8.h & (CFLAG | BITS35));
}

/*	rrca
 *	.210
 */
static void m0F(_core_o *_core)
{
	uint8 btmp = _core->af.i8.h;

	_core->af.i8.h = (btmp << 7) | (btmp >> 1);

	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | VFLAG)) |
		(btmp & CFLAG) | (_core->af.i8.h & BITS35);
}

/*	rla
 *	.209
 */
static void m17(_core_o *_core)
{
	uint8 btmp = _core->af.i8.h;

	_core->af.i8.h = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | VFLAG)) |
		(btmp >> 7) | (_core->af.i8.h & BITS35);
}

/*	rra
 *	.211
 */
static void m1F(_core_o *_core)
{
	uint8 btmp = _core->af.i8.h;

	_core->af.i8.h = (_core->af.i8.l << 7) | (btmp >> 1);

	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | VFLAG)) |
		(btmp & CFLAG) | (_core->af.i8.h & BITS35);
}

#include "daa.h"

/*	daa
 *	.184
 */
static void m27(_core_o *_core)
{
#if 0
	/* This function was taken from the M.E.S.S. Z80 emulator */

	uint8 cflag, nflag, hflag;
	uint8 lo, hi, diff;
	uint8 a, f;

	cflag = _core->af.i8.l & CFLAG;
	nflag = _core->af.i8.l & NFLAG;
	hflag = _core->af.i8.l & HFLAG;
	a = _core->af.i8.h;
	lo = a & 0x0f;
	hi = (a >> 8);

	if (cflag) {
		diff = (lo <= 9 && !hflag) ? 0x60 : 0x66;
	} else if (lo >= 10) {
		diff = (hi <= 8) ? 0x06 : 0x66;
	} else if (hi >= 10) {
		diff = (hflag) ? 0x66 : 0x60;
	} else {
		diff = (hflag) ? 0x06 : 0x00;
	}

	if (nflag)
		a -= diff;
	else
		a += diff;

	f = (a & (SFLAG | BITS35)) | (_core->af.i8.l & CNFLAG);

	if (!a)
		f |= ZFLAG;

	if (_parity(a))
		f |= VFLAG;

	if (hi >= 10)
		f |= CFLAG;
	if (lo > 9 && hi >= 9)
		f |= CFLAG;

	if (nflag) {
		if (hflag && lo <= 5)
			f |= HFLAG;
	} else if (lo >= 10)
		f |= HFLAG;

	_core->af.i8.h = a;
	_core->af.i8.l = f;
#else
	uint16 af = _core->af.i8.h;

	if (_core->af.i8.l & CFLAG)
		af |= 0x100;
	if (_core->af.i8.l & HFLAG)
		af |= 0x200;
	if (_core->af.i8.l & NFLAG)
		af |= 0x400;

	_core->af.i16 = _daa[af];
#endif
}

INTERNAL void _add_ii_d(_core_o *_core, uint16 val)
{
	uint32 res;
	uint16 hl;
	uint8 f;

	_core->addr = hl = _core->ii->i16;

	res = (uint32)hl + (uint32)val;

	f = ((res >> 8) & BITS35) | ((res >> 16) & CFLAG);

	if ((hl ^ val ^ res) & 0x1000)
		f |= HFLAG;

	_core->ii->i16 = res;
	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | VFLAG)) | f;

	_cycle(_core, 4 + 3);
}

/*	add ii, bc
 *	.197
 */
static void m09(_core_o *_core)
{
	_add_ii_d(_core, _core->bc.i16);
}

/*	add ii, de
 *	.197
 */
static void m19(_core_o *_core)
{
	_add_ii_d(_core, _core->de.i16);
}

/*	add ii, ii
 *	.197
 */
static void m29(_core_o *_core)
{
	_add_ii_d(_core, _core->ii->i16);
}

/*	add ii, sp
 *	.197
 */
static void m39(_core_o *_core)
{
	_add_ii_d(_core, _core->sp);
}

/*	cpl
 *	.186
 */
static void m2F(_core_o *_core)
{
	_core->af.i8.h = ~_core->af.i8.h;
	_core->af.i8.l |= NHFLAG;
}

/*	scf
 *	.189
 */
static void m37(_core_o *_core)
{
	_core->af.i8.l &= nhFLAG;
	_core->af.i8.l |= CFLAG;
}

/*	ccf
 *	.188
 */
static void m3F(_core_o *_core)
{
	if (_core->af.i8.l & CFLAG) {
		_core->af.i8.l &= cnFLAG;
		_core->af.i8.l |= HFLAG;
	} else {
		_core->af.i8.l &= nhFLAG;
		_core->af.i8.l |= CFLAG;
	}
}

/*	ld b, b
 *	.99
 */
static void m40(_core_o *_core)
{
	_core->bc.i8.h = _core->bc.i8.h;
}
/*	ld b, c
 *	.99
 */
static void m41(_core_o *_core)
{
	_core->bc.i8.h = _core->bc.i8.l;
}
/*	ld b, d
 *	.99
 */
static void m42(_core_o *_core)
{
	_core->bc.i8.h = _core->de.i8.h;
}
/*	ld b, e
 *	.99
 */
static void m43(_core_o *_core)
{
	_core->bc.i8.h = _core->de.i8.l;
}
/*	ld b, h
 *	.99
 */
static void m44(_core_o *_core)
{
	_core->bc.i8.h = _core->ii->i8.h;
}
/*	ld b, l
 *	.99
 */
static void m45(_core_o *_core)
{
	_core->bc.i8.h = _core->ii->i8.l;
}
/*	ld b, (ii+d)
 *	.102
 */
static void i46(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->bc.i8.h = _ld_b(_core, addr);
}
/*	ld b, (hl)
 *	.99
 */
static void m46(_core_o *_core)
{
	_core->bc.i8.h = _ld_b(_core, _core->hl.i16);
}
/*	ld b, a
 *	.99
 */
static void m47(_core_o *_core)
{
	_core->bc.i8.h = _core->af.i8.h;
}
/*	ld c, b
 *	.99
 */
static void m48(_core_o *_core)
{
	_core->bc.i8.l = _core->bc.i8.h;
}
/*	ld c, c
 *	.99
 */
static void m49(_core_o *_core)
{
	_core->bc.i8.l = _core->bc.i8.l;
}
/*	ld c, d
 *	.99
 */
static void m4A(_core_o *_core)
{
	_core->bc.i8.l = _core->de.i8.h;
}
/*	ld c, e
 *	.99
 */
static void m4B(_core_o *_core)
{
	_core->bc.i8.l = _core->de.i8.l;
}
/*	ld c, h
 *	.99
 */
static void m4C(_core_o *_core)
{
	_core->bc.i8.l = _core->ii->i8.h;
}
/*	ld c, l
 *	.99
 */
static void m4D(_core_o *_core)
{
	_core->bc.i8.l = _core->ii->i8.l;
}
/*	ld c, (ii+d)
 *	.102
 */
static void i4E(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->bc.i8.l = _ld_b(_core, addr);
}
/*	ld c, (hl)
 *	.99
 */
static void m4E(_core_o *_core)
{
	_core->bc.i8.l = _ld_b(_core, _core->hl.i16);
}
/*	ld c, a
 *	.99
 */
static void m4F(_core_o *_core)
{
	_core->bc.i8.l = _core->af.i8.h;
}

/*	ld d, b
 *	.99
 */
static void m50(_core_o *_core)
{
	_core->de.i8.h = _core->bc.i8.h;
}
/*	ld d, c
 *	.99
 */
static void m51(_core_o *_core)
{
	_core->de.i8.h = _core->bc.i8.l;
}
/*	ld d, d
 *	.99
 */
static void m52(_core_o *_core)
{
	_core->de.i8.h = _core->de.i8.h;
}
/*	ld d, e
 *	.99
 */
static void m53(_core_o *_core)
{
	_core->de.i8.h = _core->de.i8.l;
}
/*	ld d, h
 *	.99
 */
static void m54(_core_o *_core)
{
	_core->de.i8.h = _core->ii->i8.h;
}
/*	ld d, l
 *	.99
 */
static void m55(_core_o *_core)
{
	_core->de.i8.h = _core->ii->i8.l;
}
/*	ld d, (ii+d)
 *	.102
 */
static void i56(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->de.i8.h = _ld_b(_core, addr);
}
/*	ld d, (hl)
 *	.99
 */
static void m56(_core_o *_core)
{
	_core->de.i8.h = _ld_b(_core, _core->hl.i16);
}
/*	ld d, a
 *	.99
 */
static void m57(_core_o *_core)
{
	_core->de.i8.h = _core->af.i8.h;
}
/*	ld e, b
 *	.99
 */
static void m58(_core_o *_core)
{
	_core->de.i8.l = _core->bc.i8.h;
}
/*	ld e, c
 *	.99
 */
static void m59(_core_o *_core)
{
	_core->de.i8.l = _core->bc.i8.l;
}
/*	ld e, d
 *	.99
 */
static void m5A(_core_o *_core)
{
	_core->de.i8.l = _core->de.i8.h;
}
/*	ld e, e
 *	.99
 */
static void m5B(_core_o *_core)
{
	_core->de.i8.l = _core->de.i8.l;
}
/*	ld e, h
 *	.99
 */
static void m5C(_core_o *_core)
{
	_core->de.i8.l = _core->ii->i8.h;
}
/*	ld e, l
 *	.99
 */
static void m5D(_core_o *_core)
{
	_core->de.i8.l = _core->ii->i8.l;
}
/*	ld e, (ii+d)
 *	.102
 */
static void i5E(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->de.i8.l = _ld_b(_core, addr);
}
/*	ld e, (hl)
 *	.99
 */
static void m5E(_core_o *_core)
{
	_core->de.i8.l = _ld_b(_core, _core->hl.i16);
}
/*	ld e, a
 *	.99
 */
static void m5F(_core_o *_core)
{
	_core->de.i8.l = _core->af.i8.h;
}

/*	ld h, b
 *	.99
 */
static void m60(_core_o *_core)
{
	_core->ii->i8.h = _core->bc.i8.h;
}
/*	ld h, c
 *	.99
 */
static void m61(_core_o *_core)
{
	_core->ii->i8.h = _core->bc.i8.l;
}
/*	ld h, d
 *	.99
 */
static void m62(_core_o *_core)
{
	_core->ii->i8.h = _core->de.i8.h;
}
/*	ld h, e
 *	.99
 */
static void m63(_core_o *_core)
{
	_core->ii->i8.h = _core->de.i8.l;
}
/*	ld h, h
 *	.99
 */
static void m64(_core_o *_core)
{
	_core->ii->i8.h = _core->ii->i8.h;
}
/*	ld h, l
 *	.99
 */
static void m65(_core_o *_core)
{
	_core->ii->i8.h = _core->ii->i8.l;
}
/*	ld h, (ii+d)
 *	.102
 */
static void i66(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->hl.i8.h = _ld_b(_core, addr);
}
/*	ld h, (hl)
 *	.99
 */
static void m66(_core_o *_core)
{
	_core->hl.i8.h = _ld_b(_core, _core->hl.i16);
}
/*	ld h, a
 *	.99
 */
static void m67(_core_o *_core)
{
	_core->ii->i8.h = _core->af.i8.h;
}
/*	ld l, b
 *	.99
 */
static void m68(_core_o *_core)
{
	_core->ii->i8.l = _core->bc.i8.h;
}
/*	ld l, c
 *	.99
 */
static void m69(_core_o *_core)
{
	_core->ii->i8.l = _core->bc.i8.l;
}
/*	ld l, d
 *	.99
 */
static void m6A(_core_o *_core)
{
	_core->ii->i8.l = _core->de.i8.h;
}
/*	ld l, e
 *	.99
 */
static void m6B(_core_o *_core)
{
	_core->ii->i8.l = _core->de.i8.l;
}
/*	ld l, h
 *	.99
 */
static void m6C(_core_o *_core)
{
	_core->ii->i8.l = _core->ii->i8.h;
}
/*	ld l, l
 *	.99
 */
static void m6D(_core_o *_core)
{
	_core->ii->i8.l = _core->ii->i8.l;
}
/*	ld l, (ii+d)
 *	.102
 */
static void i6E(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->hl.i8.l = _ld_b(_core, addr);
}
/*	ld l, (hl)
 *	.99
 */
static void m6E(_core_o *_core)
{
	_core->hl.i8.l = _ld_b(_core, _core->hl.i16);
}
/*	ld l, a
 *	.99
 */
static void m6F(_core_o *_core)
{
	_core->ii->i8.l = _core->af.i8.h;
}

/*	ld (ii+d), b
 *	.105
 */
static void i70(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->bc.i8.h);
}
/*	ld (ii+d), c
 *	.105
 */
static void i71(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->bc.i8.l);
}
/*	ld (ii+d), d
 *	.105
 */
static void i72(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->de.i8.h);
}
/*	ld (ii+d), e
 *	.105
 */
static void i73(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->de.i8.l);
}
/*	ld (ii+d), h
 *	.105
 */
static void i74(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->hl.i8.h);
}
/*	ld (ii+d), l
 *	.105
 */
static void i75(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->hl.i8.l);
}
/*	ld (ii+d), a
 *	.105
 */
static void i77(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_st_b(_core, addr, _core->af.i8.h);
}

/*	ld (hl), b
 *	.104
 */
static void m70(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->bc.i8.h);
}
/*	ld (hl), c
 *	.104
 */
static void m71(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->bc.i8.l);
}
/*	ld (hl), d
 *	.104
 */
static void m72(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->de.i8.h);
}
/*	ld (hl), e
 *	.104
 */
static void m73(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->de.i8.l);
}
/*	ld (hl), h
 *	.104
 */
static void m74(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->hl.i8.h);
}
/*	ld (hl), l
 *	.104
 */
static void m75(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->hl.i8.l);
}
/*	ld (hl), a
 *	.104
 */
static void m77(_core_o *_core)
{
	_st_b(_core, _core->hl.i16, _core->af.i8.h);
}
/*	ld a, b
 *	.99
 */
static void m78(_core_o *_core)
{
	_core->af.i8.h = _core->bc.i8.h;
}
/*	ld a, c
 *	.99
 */
static void m79(_core_o *_core)
{
	_core->af.i8.h = _core->bc.i8.l;
}
/*	ld a, d
 *	.99
 */
static void m7A(_core_o *_core)
{
	_core->af.i8.h = _core->de.i8.h;
}
/*	ld a, e
 *	.99
 */
static void m7B(_core_o *_core)
{
	_core->af.i8.h = _core->de.i8.l;
}
/*	ld a, h
 *	.99
 */
static void m7C(_core_o *_core)
{
	_core->af.i8.h = _core->ii->i8.h;
}
/*	ld a, l
 *	.99
 */
static void m7D(_core_o *_core)
{
	_core->af.i8.h = _core->ii->i8.l;
}
/*	ld a, (ii+d)
 *	.102
 */
static void i7E(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->af.i8.h = _ld_b(_core, addr);
}
/*	ld a, (hl)
 *	.99
 */
static void m7E(_core_o *_core)
{
	_core->af.i8.h = _ld_b(_core, _core->hl.i16);
}
/*	ld a, a
 *	.99
 */
static void m7F(_core_o *_core)
{
	_core->af.i8.h = _core->af.i8.h;
}

/*	halt
 *	.191
 */
static void m76(_core_o *_core)
{
	_core->pc--;

	_core->halted = 1;
}

INTERNAL uint8 _add(_core_o *_core, uint8 val, uint8 with_carry)
{
	uint8 a, f;
	uint16 res;

	a = _core->af.i8.h;

	res = (uint16)a + (uint16)val + (with_carry & _core->af.i8.l);

	f = (res & (SFLAG | BITS35)) | ((res >> 8) & CFLAG);

	if (!(res & 0xff))
		f |= ZFLAG;

	f |= ((a ^ val ^ res) & HFLAG);

	if (~(a ^ val) & (a ^ res) & 0x80)
		f |= VFLAG;

	_core->af.i8.l = f;
	return res;
}

INTERNAL uint8 _sub(_core_o *_core, uint8 val, uint8 with_borrow)
{
	uint8 a, f;
	uint16 res;

	a = _core->af.i8.h;

	res = (uint16)a - (uint16)val - (with_borrow & _core->af.i8.l);

	f = (res & (SFLAG | BITS35)) | ((res >> 8) & CFLAG);

	if (!(res & 0xff))
		f |= ZFLAG;

	f |= ((a ^ val ^ res) & HFLAG);

	if ((a ^ val) & (a ^ res) & 0x80)
		f |= VFLAG;

	_core->af.i8.l = f | NFLAG;
	return res;
}

INTERNAL void _f_logic(_core_o *_core, uint8 hflag)
{
	uint8 val;
	uint8 f;

	val = _core->af.i8.h;

	f = (val & (SFLAG | BITS35)) | hflag;

	if (!val)
		f |= ZFLAG;

	if (_parity(val))
		f |= VFLAG;

	_core->af.i8.l = f;
}

/*	add a, b
 *	.158
 */
static void m80(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->bc.i8.h, 0);
}
/*	add a, c
 *	.158
 */
static void m81(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->bc.i8.l, 0);
}
/*	add a, d
 *	.158
 */
static void m82(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->de.i8.h, 0);
}
/*	add a, e
 *	.158
 */
static void m83(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->de.i8.l, 0);
}
/*	add a, h
 *	.158
 */
static void m84(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->ii->i8.h, 0);
}
/*	add a, l
 *	.158
 */
static void m85(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->ii->i8.l, 0);
}
/*	add a, (ii+d)
 *	.158
 */
static void i86(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);
	btmp = _ld_b(_core, addr);
	_core->af.i8.h = _add(_core, btmp, 0);
}
/*	add a, (hl)
 *	.158
 */
static void m86(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_core->af.i8.h = _add(_core, btmp, 0);
}
/*	add a, a
 *	.158
 */
static void m87(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->af.i8.h, 0);
}
/*	add a, n
 *	.160
 */
static void mC6(_core_o *_core)
{
	uint8 btmp;

	btmp = _imm_b(_core);
	_core->af.i8.h = _add(_core, btmp, 0);
}

/*	adc a, b
 *	.164
 */
static void m88(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->bc.i8.h, CFLAG);
}
/*	adc a, c
 *	.164
 */
static void m89(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->bc.i8.l, CFLAG);
}
/*	adc a, d
 *	.164
 */
static void m8A(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->de.i8.h, CFLAG);
}
/*	adc a, e
 *	.164
 */
static void m8B(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->de.i8.l, CFLAG);
}
/*	adc a, h
 *	.164
 */
static void m8C(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->ii->i8.h, CFLAG);
}
/*	adc a, l
 *	.164
 */
static void m8D(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->ii->i8.l, CFLAG);
}
/*	adc a, (ii+d)
 *	.164
 */
static void i8E(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);
	btmp = _ld_b(_core, addr);
	_core->af.i8.h = _add(_core, btmp, CFLAG);
}
/*	adc a, (hl)
 *	.164
 */
static void m8E(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_core->af.i8.h = _add(_core, btmp, CFLAG);
}
/*	adc a, a
 *	.164
 */
static void m8F(_core_o *_core)
{
	_core->af.i8.h = _add(_core, _core->af.i8.h, CFLAG);
}
/*	adc a, n
 *	.164
 */
static void mCE(_core_o *_core)
{
	uint8 btmp;

	btmp = _imm_b(_core);
	_core->af.i8.h = _add(_core, btmp, CFLAG);
}

/*	sub a, b
 *	.166
 */
static void m90(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->bc.i8.h, 0);
}
/*	sub a, c
 *	.166
 */
static void m91(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->bc.i8.l, 0);
}
/*	sub a, d
 *	.166
 */
static void m92(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->de.i8.h, 0);
}
/*	sub a, e
 *	.166
 */
static void m93(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->de.i8.l, 0);
}
/*	sub a, h
 *	.166
 */
static void m94(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->ii->i8.h, 0);
}
/*	sub a, l
 *	.166
 */
static void m95(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->ii->i8.l, 0);
}
/*	sub a, (ii+d)
 *	.166
 */
static void i96(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);
	btmp = _ld_b(_core, addr);
	_core->af.i8.h = _sub(_core, btmp, 0);
}
/*	sub a, (hl)
 *	.166
 */
static void m96(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_core->af.i8.h = _sub(_core, btmp, 0);
}
/*	sub a, a
 *	.166
 */
static void m97(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->af.i8.h, 0);
}
/*	sub a, n
 *	.166
 */
static void mD6(_core_o *_core)
{
	uint8 btmp;

	btmp = _imm_b(_core);
	_core->af.i8.h = _sub(_core, btmp, 0);
}

/*	sbc a, b
 *	.168
 */
static void m98(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->bc.i8.h, CFLAG);
}
/*	sbc a, c
 *	.168
 */
static void m99(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->bc.i8.l, CFLAG);
}
/*	sbc a, d
 *	.168
 */
static void m9A(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->de.i8.h, CFLAG);
}
/*	sbc a, e
 *	.168
 */
static void m9B(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->de.i8.l, CFLAG);
}
/*	sbc a, h
 *	.168
 */
static void m9C(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->ii->i8.h, CFLAG);
}
/*	sbc a, l
 *	.168
 */
static void m9D(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->ii->i8.l, CFLAG);
}
/*	sbc a, (ii+d)
 *	.168
 */
static void i9E(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);
	btmp = _ld_b(_core, addr);
	_core->af.i8.h = _sub(_core, btmp, CFLAG);
}
/*	sbc a, (hl)
 *	.168
 */
static void m9E(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_core->af.i8.h = _sub(_core, btmp, CFLAG);
}
/*	sbc a, a
 *	.168
 */
static void m9F(_core_o *_core)
{
	_core->af.i8.h = _sub(_core, _core->af.i8.h, CFLAG);
}
/*	sbc a, n
 *	.168
 */
static void mDE(_core_o *_core)
{
	uint8 btmp;

	btmp = _imm_b(_core);
	_core->af.i8.h = _sub(_core, btmp, CFLAG);
}

/*	and a, b
 *	.170
 */
static void mA0(_core_o *_core)
{
	_core->af.i8.h &= _core->bc.i8.h;
	_f_logic(_core, HFLAG);
}
/*	and a, c
 *	.170
 */
static void mA1(_core_o *_core)
{
	_core->af.i8.h &= _core->bc.i8.l;
	_f_logic(_core, HFLAG);
}
/*	and a, d
 *	.170
 */
static void mA2(_core_o *_core)
{
	_core->af.i8.h &= _core->de.i8.h;
	_f_logic(_core, HFLAG);
}
/*	and a, e
 *	.170
 */
static void mA3(_core_o *_core)
{
	_core->af.i8.h &= _core->de.i8.l;
	_f_logic(_core, HFLAG);
}
/*	and a, h
 *	.170
 */
static void mA4(_core_o *_core)
{
	_core->af.i8.h &= _core->ii->i8.h;
	_f_logic(_core, HFLAG);
}
/*	and a, l
 *	.170
 */
static void mA5(_core_o *_core)
{
	_core->af.i8.h &= _core->ii->i8.l;
	_f_logic(_core, HFLAG);
}
/*	and a, (ii+d)
 *	.170
 */
static void iA6(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->af.i8.h &= _ld_b(_core, addr);
	_f_logic(_core, HFLAG);
}
/*	and a, (hl)
 *	.170
 */
static void mA6(_core_o *_core)
{
	_core->af.i8.h &= _ld_b(_core, _core->hl.i16);
	_f_logic(_core, HFLAG);
}
/*	and a, a
 *	.170
 */
static void mA7(_core_o *_core)
{
	_core->af.i8.h &= _core->af.i8.h;
	_f_logic(_core, HFLAG);
}
/*	and a, n
 *	.170
 */
static void mE6(_core_o *_core)
{
	_core->af.i8.h &= _imm_b(_core);
	_f_logic(_core, HFLAG);
}

/*	xor a, b
 *	.170
 */
static void mA8(_core_o *_core)
{
	_core->af.i8.h ^= _core->bc.i8.h;
	_f_logic(_core, 0);
}
/*	xor a, c
 *	.170
 */
static void mA9(_core_o *_core)
{
	_core->af.i8.h ^= _core->bc.i8.l;
	_f_logic(_core, 0);
}
/*	xor a, d
 *	.170
 */
static void mAA(_core_o *_core)
{
	_core->af.i8.h ^= _core->de.i8.h;
	_f_logic(_core, 0);
}
/*	xor a, e
 *	.170
 */
static void mAB(_core_o *_core)
{
	_core->af.i8.h ^= _core->de.i8.l;
	_f_logic(_core, 0);
}
/*	xor a, h
 *	.170
 */
static void mAC(_core_o *_core)
{
	_core->af.i8.h ^= _core->ii->i8.h;
	_f_logic(_core, 0);
}
/*	xor a, l
 *	.170
 */
static void mAD(_core_o *_core)
{
	_core->af.i8.h ^= _core->ii->i8.l;
	_f_logic(_core, 0);
}
/*	xor a, (ii+d)
 *	.170
 */
static void iAE(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->af.i8.h ^= _ld_b(_core, addr);
	_f_logic(_core, 0);
}
/*	xor a, (hl)
 *	.170
 */
static void mAE(_core_o *_core)
{
	_core->af.i8.h ^= _ld_b(_core, _core->hl.i16);
	_f_logic(_core, 0);
}
/*	xor a, a
 *	.170
 */
static void mAF(_core_o *_core)
{
	_core->af.i8.h ^= _core->af.i8.h;
	_f_logic(_core, 0);
}
/*	xor a, n
 *	.170
 */
static void mEE(_core_o *_core)
{
	_core->af.i8.h ^= _imm_b(_core);
	_f_logic(_core, 0);
}

/*	or a, b
 *	.170
 */
static void mB0(_core_o *_core)
{
	_core->af.i8.h |= _core->bc.i8.h;
	_f_logic(_core, 0);
}
/*	or a, c
 *	.170
 */
static void mB1(_core_o *_core)
{
	_core->af.i8.h |= _core->bc.i8.l;
	_f_logic(_core, 0);
}
/*	or a, d
 *	.170
 */
static void mB2(_core_o *_core)
{
	_core->af.i8.h |= _core->de.i8.h;
	_f_logic(_core, 0);
}
/*	or a, e
 *	.170
 */
static void mB3(_core_o *_core)
{
	_core->af.i8.h |= _core->de.i8.l;
	_f_logic(_core, 0);
}
/*	or a, h
 *	.170
 */
static void mB4(_core_o *_core)
{
	_core->af.i8.h |= _core->ii->i8.h;
	_f_logic(_core, 0);
}
/*	or a, l
 *	.170
 */
static void mB5(_core_o *_core)
{
	_core->af.i8.h |= _core->ii->i8.l;
	_f_logic(_core, 0);
}
/*	or a, (ii+d)
 *	.170
 */
static void iB6(_core_o *_core)
{
	uint16 addr;

	addr = _addr_ii(_core);
	_core->af.i8.h |= _ld_b(_core, addr);
	_f_logic(_core, 0);
}
/*	or a, (hl)
 *	.170
 */
static void mB6(_core_o *_core)
{
	_core->af.i8.h |= _ld_b(_core, _core->hl.i16);
	_f_logic(_core, 0);
}
/*	or a, a
 *	.170
 */
static void mB7(_core_o *_core)
{
	_core->af.i8.h |= _core->af.i8.h;
	_f_logic(_core, 0);
}
/*	or a, n
 *	.170
 */
static void mF6(_core_o *_core)
{
	_core->af.i8.h |= _imm_b(_core);
	_f_logic(_core, 0);
}

/*	cp a, b
 *	.170
 */
static void mB8(_core_o *_core)
{
	_sub(_core, _core->bc.i8.h, 0);
}
/*	cp a, c
 *	.170
 */
static void mB9(_core_o *_core)
{
	_sub(_core, _core->bc.i8.l, 0);
}
/*	cp a, d
 *	.170
 */
static void mBA(_core_o *_core)
{
	_sub(_core, _core->de.i8.h, 0);
}
/*	cp a, e
 *	.170
 */
static void mBB(_core_o *_core)
{
	_sub(_core, _core->de.i8.l, 0);
}
/*	cp a, h
 *	.170
 */
static void mBC(_core_o *_core)
{
	_sub(_core, _core->ii->i8.h, 0);
}
/*	cp a, l
 *	.170
 */
static void mBD(_core_o *_core)
{
	_sub(_core, _core->ii->i8.l, 0);
}
/*	cp a, (ii+d)
 *	.170
 */
static void iBE(_core_o *_core)
{
	uint16 addr;
	uint8 btmp;

	addr = _addr_ii(_core);
	btmp = _ld_b(_core, addr);
	_sub(_core, btmp, 0);
}
/*	cp a, (hl)
 *	.170
 */
static void mBE(_core_o *_core)
{
	uint8 btmp;

	btmp = _ld_b(_core, _core->hl.i16);
	_sub(_core, btmp, 0);
}
/*	cp a, a
 *	.170
 */
static void mBF(_core_o *_core)
{
	_sub(_core, _core->af.i8.h, 0);
}
/*	cp a, n
 *	.170
 */
static void mFE(_core_o *_core)
{
	uint8 btmp;

	btmp = _imm_b(_core);
	_sub(_core, btmp, 0);
}

INTERNAL uint16 _pop(_core_o *_core)
{
	uint16 res;

	res = _ld_w(_core, _core->sp);
	_core->sp += 2;

	return res;
}

INTERNAL void _ret(_core_o *_core)
{
	_core->pc = _pop(_core);
}

INTERNAL void _ret_f(_core_o *_core, uint8 flag)
{
	_cycle(_core, 1);

	if (_core->af.i8.l & flag)
		_ret(_core);
}

INTERNAL void _ret_nf(_core_o *_core, uint8 flag)
{
	_cycle(_core, 1);

	if (!(_core->af.i8.l & flag))
		_ret(_core);
}

/*	ret
 *	.278
 */
static void mC9(_core_o *_core)
{
	_ret(_core);
}

/*	ret nz
 *	.279
 */
static void mC0(_core_o *_core)
{
	_ret_nf(_core, ZFLAG);
}
/*	ret z
 *	.279
 */
static void mC8(_core_o *_core)
{
	_ret_f(_core, ZFLAG);
}
/*	ret nc
 *	.279
 */
static void mD0(_core_o *_core)
{
	_ret_nf(_core, CFLAG);
}
/*	ret c
 *	.279
 */
static void mD8(_core_o *_core)
{
	_ret_f(_core, CFLAG);
}
/*	ret po
 *	.279
 */
static void mE0(_core_o *_core)
{
	_ret_nf(_core, VFLAG);
}
/*	ret pe
 *	.279
 */
static void mE8(_core_o *_core)
{
	_ret_f(_core, VFLAG);
}
/*	ret p 
 *	.279
 */
static void mF0(_core_o *_core)
{
	_ret_nf(_core, SFLAG);
}
/*	ret m
 *	.279
 */
static void mF8(_core_o *_core)
{
	_ret_f(_core, SFLAG);
}

/*	pop bc
 *	.137
 */
static void mC1(_core_o *_core)
{
	_core->bc.i16 = _pop(_core);
}
/*	pop de
 *	.137
 */
static void mD1(_core_o *_core)
{
	_core->de.i16 = _pop(_core);
}
/*	pop ii
 *	.137
 */
static void mE1(_core_o *_core)
{
	_core->ii->i16 = _pop(_core);
}
/*	pop af
 *	.137
 */
static void mF1(_core_o *_core)
{
	_core->af.i16 = _pop(_core);
}

INTERNAL void _jp(_core_o *_core, uint16 addr)
{
	_core->pc = addr;
}

INTERNAL void _jp_f(_core_o *_core, uint8 flag)
{
	uint16 addr = _imm_w(_core);

	if (_core->af.i8.l & flag)
		_jp(_core, addr);
}

INTERNAL void _jp_nf(_core_o *_core, uint8 flag)
{
	uint16 addr = _imm_w(_core);

	if (!(_core->af.i8.l & flag))
		_jp(_core, addr);
}

/*	jp nn
 *	.256
 */
static void mC3(_core_o *_core)
{
	uint16 addr = _imm_w(_core);

	_jp(_core, addr);
}

/*	jp nz, nn
 *	.257
 */
static void mC2(_core_o *_core)
{
	_jp_nf(_core, ZFLAG);
}
/*	jp z, nn
 *	.257
 */
static void mCA(_core_o *_core)
{
	_jp_f(_core, ZFLAG);
}
/*	jp nc, nn
 *	.257
 */
static void mD2(_core_o *_core)
{
	_jp_nf(_core, CFLAG);
}
/*	jp c, nn
 *	.257
 */
static void mDA(_core_o *_core)
{
	_jp_f(_core, CFLAG);
}
/*	jp po, nn
 *	.257
 */
static void mE2(_core_o *_core)
{
	_jp_nf(_core, VFLAG);
}
/*	jp pe, nn
 *	.257
 */
static void mEA(_core_o *_core)
{
	_jp_f(_core, VFLAG);
}
/*	jp p, nn
 *	.257
 */
static void mF2(_core_o *_core)
{
	_jp_nf(_core, SFLAG);
}
/*	jp m, nn
 *	.257
 */
static void mFA(_core_o *_core)
{
	_jp_f(_core, SFLAG);
}

INTERNAL void _out(_core_o *_core, uint16 addr, uint8 data)
{
	_cycle(_core, 4);
	dev_out(_core->dev, addr, data);
}

INTERNAL uint8 _in(_core_o *_core, uint16 addr)
{
	_cycle(_core, 4);
	return dev_in(_core->dev, addr);
}

/*	out (n), a
 *	.297
 */
static void mD3(_core_o *_core)
{
	uint16 addr;

	addr = (_core->af.i8.h << 8) | _imm_b(_core);
	_out(_core, addr, _core->af.i8.h);
}

/*	in a, (n)
 *	.287
 */
static void mDB(_core_o *_core)
{
	uint16 addr;

	addr = (_core->af.i8.h << 8) | _imm_b(_core);
	_core->af.i8.h = _in(_core, addr);
}

/*	ex (sp), ii
 *	.143
 */
static void mE3(_core_o *_core)
{
	uint16 wtmp;

	wtmp = _ld_w(_core, _core->sp);
	_st_w(_core, _core->sp, _core->ii->i16);
	_core->ii->i16 = wtmp;

	_cycle(_core, 3);
}

/*	ex de, hl
 *	.140
 */
static void mEB(_core_o *_core)
{
	uint16 wtmp;

	wtmp = _core->hl.i16;
	_core->hl.i16 = _core->de.i16;
	_core->de.i16 = wtmp;
}

/*	di
 *	.192
 */
static void mF3(_core_o *_core)
{
	_core->iff1 = 0;
	_core->iff2 = 0;
}

/*	ei
 *	.193
 */
static void mFB(_core_o *_core)
{
	_core->iff1 = 1;
	_core->iff2 = 1;

	_core->active = 2;
}

INTERNAL void _push(_core_o *_core, uint16 data)
{
	_core->sp -= 2;
	_cycle(_core, 1);

	_st_w(_core, _core->sp, data);
}

INTERNAL void _call(_core_o *_core, uint16 addr)
{
	_push(_core, _core->pc);
	_core->pc = addr;
}

INTERNAL void _call_f(_core_o *_core, uint8 flag)
{
	uint16 addr;
	addr = _imm_w(_core);

	if (_core->af.i8.l & flag)
		_call(_core, addr);
}

INTERNAL void _call_nf(_core_o *_core, int flag)
{
	uint16 addr;
	addr = _imm_w(_core);

	if (!(_core->af.i8.l & flag))
		_call(_core, addr);
}

/*	call nn
 *	.273
 */
static void mCD(_core_o *_core)
{
	uint16 addr;
	addr = _imm_w(_core);

	_call(_core, addr);
}

/*	call nz, nn
 *	.275
 */
static void mC4(_core_o *_core)
{
	_call_nf(_core, ZFLAG);
}
/*	call z, nn
 *	.275
 */
static void mCC(_core_o *_core)
{
	_call_f(_core, ZFLAG);
}
/*	call nc, nn
 *	.275
 */
static void mD4(_core_o *_core)
{
	_call_nf(_core, CFLAG);
}
/*	call c, nn
 *	.275
 */
static void mDC(_core_o *_core)
{
	_call_f(_core, CFLAG);
}
/*	call po, nn
 *	.275
 */
static void mE4(_core_o *_core)
{
	_call_nf(_core, VFLAG);
}
/*	call pe, nn
 *	.275
 */
static void mEC(_core_o *_core)
{
	_call_f(_core, VFLAG);
}
/*	call p, nn
 *	.275
 */
static void mF4(_core_o *_core)
{
	_call_nf(_core, SFLAG);
}
/*	call m, nn
 *	.275
 */
static void mFC(_core_o *_core)
{
	_call_f(_core, SFLAG);
}

/*	push bc
 *	.134
 */
static void mC5(_core_o *_core)
{
	_push(_core, _core->bc.i16);
}
/*	push de
 *	.134
 */
static void mD5(_core_o *_core)
{
	_push(_core, _core->de.i16);
}
/*	push ii
 *	.134
 */
static void mE5(_core_o *_core)
{
	_push(_core, _core->ii->i16);
}
/*	push af
 *	.134
 */
static void mF5(_core_o *_core)
{
	_push(_core, _core->af.i16);
}

/*	rst 00h
 *	.285
 */
static void mC7(_core_o *_core)
{
	_call(_core, 0x00);
}
/*	rst 08h
 *	.285
 */
static void mCF(_core_o *_core)
{
	_call(_core, 0x08);
}
/*	rst 10h
 *	.285
 */
static void mD7(_core_o *_core)
{
	_call(_core, 0x10);
}
/*	rst 18h
 *	.285
 */
static void mDF(_core_o *_core)
{
	_call(_core, 0x18);
}
/*	rst 20h
 *	.285
 */
static void mE7(_core_o *_core)
{
	_call(_core, 0x20);
}
/*	rst 28h
 *	.285
 */
static void mEF(_core_o *_core)
{
	_call(_core, 0x28);
}
/*	rst 30h
 *	.285
 */
static void mF7(_core_o *_core)
{
	_call(_core, 0x30);
}
/*	rst 38h
 *	.285
 */
static void mFF(_core_o *_core)
{
	_call(_core, 0x38);
}

/*	exx
 *	.142
 */
static void mD9(_core_o *_core)
{
	uint16 wtmp;

	wtmp = _core->bc.i16;
	_core->bc.i16 = _core->bc_;
	_core->bc_ = wtmp;

	wtmp = _core->de.i16;
	_core->de.i16 = _core->de_;
	_core->de_ = wtmp;

	wtmp = _core->hl.i16;
	_core->hl.i16 = _core->hl_;
	_core->hl_ = wtmp;
}

/*	jp (ii)
 *	.268
 */
static void mE9(_core_o *_core)
{
	_core->pc = _core->ii->i16;
}

/*	ld sp, ii
 *	.131
 */
static void mF9(_core_o *_core)
{
	_core->sp = _core->ii->i16;
}

static void iCB(_core_o *_core);
static void mCB(_core_o *_core);
static void mED(_core_o *_core);
static void mDD(_core_o *_core);
static void mFD(_core_o *_core);

INTERNAL void _f_rot(_core_o *_core, uint8 val, uint8 carry)
{
	uint8 f;

	f = (val & (SFLAG | BITS35)) | carry;

	if (!val)
		f |= ZFLAG;

	if (_parity(val))
		f |= VFLAG;

	_core->af.i8.l = f;
}

/*	rlc b
 *	.212
 */
static void a00(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->bc.i8.h, (btmp >> 7));
}
/*	rlc c
 *	.212
 */
static void a01(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->bc.i8.l, (btmp >> 7));
}
/*	rlc d
 *	.212
 */
static void a02(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->de.i8.h, (btmp >> 7));
}
/*	rlc e
 *	.212
 */
static void a03(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->de.i8.l, (btmp >> 7));
}
/*	rlc h
 *	.212
 */
static void a04(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->hl.i8.h, (btmp >> 7));
}
/*	rlc l
 *	.212
 */
static void a05(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->hl.i8.l, (btmp >> 7));
}
/*	rlc (ii+d)
 *	.212
 */
static void ia00(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, val, (btmp >> 7));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	rlc (hl)
 *	.212
 */
static void a06(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, val, (btmp >> 7));
	_st_b(_core, _core->hl.i16, val);
}
/*	rlc a
 *	.212
 */
static void a07(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp << 1) | (btmp >> 7);

	_f_rot(_core, _core->af.i8.h, (btmp >> 7));
}

/*	rrc b
 *	.223
 */
static void a08(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->bc.i8.h, (btmp & 0x01));
}
/*	rrc c
 *	.223
 */
static void a09(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->bc.i8.l, (btmp & 0x01));
}
/*	rrc d
 *	.223
 */
static void a0A(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->de.i8.h, (btmp & 0x01));
}
/*	rrc e
 *	.223
 */
static void a0B(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->de.i8.l, (btmp & 0x01));
}
/*	rrc h
 *	.223
 */
static void a0C(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->hl.i8.h, (btmp & 0x01));
}
/*	rrc l
 *	.223
 */
static void a0D(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->hl.i8.l, (btmp & 0x01));
}
/*	rrc (ii+d)
 *	.223
 */
static void ia08(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, val, (btmp & 0x01));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	rrc (hl)
 *	.223
 */
static void a0E(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, val, (btmp & 0x1));
	_st_b(_core, _core->hl.i16, val);
}
/*	rrc a
 *	.223
 */
static void a0F(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp << 7) | (btmp >> 1);

	_f_rot(_core, _core->af.i8.h, (btmp & 0x01));
}

/*	rl b
 *	.220
 */
static void a10(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->bc.i8.h, (btmp >> 7));
}
/*	rl c
 *	.220
 */
static void a11(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->bc.i8.l, (btmp >> 7));
}
/*	rl d
 *	.220
 */
static void a12(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->de.i8.h, (btmp >> 7));
}
/*	rl e
 *	.220
 */
static void a13(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->de.i8.l, (btmp >> 7));
}
/*	rl h
 *	.220
 */
static void a14(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->hl.i8.h, (btmp >> 7));
}
/*	rl l
 *	.220
 */
static void a15(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->hl.i8.l, (btmp >> 7));
}
/*	rl (ii+d)
 *	.220
 */
static void ia10(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, val, (btmp >> 7));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	rl (hl)
 *	.220
 */
static void a16(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, val, (btmp >> 7));
	_st_b(_core, _core->hl.i16, val);
}
/*	rl a
 *	.220
 */
static void a17(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp << 1) | (_core->af.i8.l & CFLAG);

	_f_rot(_core, _core->af.i8.h, (btmp >> 7));
}

/*	rr b
 *	.226
 */
static void a18(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->bc.i8.h, (btmp & 0x01));
}
/*	rr c
 *	.226
 */
static void a19(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->bc.i8.l, (btmp & 0x01));
}
/*	rr d
 *	.226
 */
static void a1A(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->de.i8.h, (btmp & 0x01));
}
/*	rr e
 *	.226
 */
static void a1B(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->de.i8.l, (btmp & 0x01));
}
/*	rr h
 *	.226
 */
static void a1C(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->hl.i8.h, (btmp & 0x01));
}
/*	rr l
 *	.226
 */
static void a1D(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->hl.i8.l, (btmp & 0x01));
}
/*	rr (ii+d)
 *	.226
 */
static void ia18(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, val, (btmp & 0x01));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	rr (hl)
 *	.226
 */
static void a1E(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, val, (btmp & 0x01));
	_st_b(_core, _core->hl.i16, val);
}
/*	rr a
 *	.226
 */
static void a1F(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (_core->af.i8.l << 7) | (btmp >> 1);

	_f_rot(_core, _core->af.i8.h, (btmp & 0x01));
}

INTERNAL void _f_shift(_core_o *_core, uint8 val, uint8 carry)
{
	uint8 f;

	f = (val & (SFLAG | BITS35)) | carry;

	if (!val)
		f |= ZFLAG;

	if (_parity(val))
		f |= VFLAG;

	_core->af.i8.l = f;
}

/*	sla b
 *	.229
 */
static void a20(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp << 1);

	_f_shift(_core, _core->bc.i8.h, (btmp >> 7));
}
/*	sla c
 *	.229
 */
static void a21(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp << 1);

	_f_shift(_core, _core->bc.i8.l, (btmp >> 7));
}
/*	sla d
 *	.229
 */
static void a22(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp << 1);

	_f_shift(_core, _core->de.i8.h, (btmp >> 7));
}
/*	sla e
 *	.229
 */
static void a23(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp << 1);

	_f_shift(_core, _core->de.i8.l, (btmp >> 7));
}
/*	sla h
 *	.229
 */
static void a24(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp << 1);

	_f_shift(_core, _core->hl.i8.h, (btmp >> 7));
}
/*	sla l
 *	.229
 */
static void a25(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp << 1);

	_f_shift(_core, _core->hl.i8.l, (btmp >> 7));
}
/*	sla (ii+d)
 *	.229
 */
static void ia20(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp << 1);

	_f_shift(_core, val, (btmp >> 7));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	sla (hl)
 *	.229
 */
static void a26(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp << 1);

	_f_shift(_core, val, (btmp >> 7));
	_st_b(_core, _core->hl.i16, val);
}
/*	sla a
 *	.229
 */
static void a27(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp << 1);

	_f_shift(_core, _core->af.i8.h, (btmp >> 7));
}

/*	sra b
 *	.232
 */
static void a28(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp >> 1);

	_f_shift(_core, _core->bc.i8.h, (btmp & 0x01));
}
/*	sra c
 *	.232
 */
static void a29(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp >> 1);

	_f_shift(_core, _core->bc.i8.l, (btmp & 0x01));
}
/*	sra d
 *	.232
 */
static void a2A(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp >> 1);

	_f_shift(_core, _core->de.i8.h, (btmp & 0x01));
}
/*	sra e
 *	.232
 */
static void a2B(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp >> 1);

	_f_shift(_core, _core->de.i8.l, (btmp & 0x01));
}
/*	sra h
 *	.232
 */
static void a2C(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp >> 1);

	_f_shift(_core, _core->hl.i8.h, (btmp & 0x01));
}
/*	sra l
 *	.232
 */
static void a2D(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp >> 1);

	_f_shift(_core, _core->hl.i8.l, (btmp & 0x01));
}
/*	sra (ii+d)
 *	.232
 */
static void ia28(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	sint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp >> 1);

	_f_shift(_core, val, (btmp & 0x01));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	sra (hl)
 *	.232
 */
static void a2E(_core_o *_core)
{
	sint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp >> 1);

	_f_shift(_core, val, (btmp & 0x1));
	_st_b(_core, _core->hl.i16, val);
}
/*	sra a
 *	.232
 */
static void a2F(_core_o *_core)
{
	sint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp >> 1);

	_f_shift(_core, _core->af.i8.h, (btmp & 0x01));
}

/*	sll b
 *	.xxx
 */
static void a30(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->bc.i8.h, (btmp >> 7));
}
/*	sll c
 *	.xxx
 */
static void a31(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->bc.i8.l, (btmp >> 7));
}
/*	sll d
 *	.xxx
 */
static void a32(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->de.i8.h, (btmp >> 7));
}
/*	sll e
 *	.xxx
 */
static void a33(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->de.i8.l, (btmp >> 7));
}
/*	sll h
 *	.xxx
 */
static void a34(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->hl.i8.h, (btmp >> 7));
}
/*	sll l
 *	.xxx
 */
static void a35(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->hl.i8.l, (btmp >> 7));
}
/*	sll (ii+d)
 *	.xxx
 */
static void ia30(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp << 1) | CFLAG;

	_f_shift(_core, val, (btmp >> 7));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	sll (hl)
 *	.xxx
 */
static void a36(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp << 1) | CFLAG;

	_f_shift(_core, val, (btmp >> 7));
	_st_b(_core, _core->hl.i16, val);
}
/*	sll a
 *	.xxx
 */
static void a37(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp << 1) | CFLAG;

	_f_shift(_core, _core->af.i8.h, (btmp >> 7));
}

/*	srl b
 *	.235
 */
static void a38(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.h;
	_core->bc.i8.h = (btmp >> 1);

	_f_shift(_core, _core->bc.i8.h, (btmp & 0x01));
}
/*	srl c
 *	.235
 */
static void a39(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->bc.i8.l;
	_core->bc.i8.l = (btmp >> 1);

	_f_shift(_core, _core->bc.i8.l, (btmp & 0x01));
}
/*	srl d
 *	.235
 */
static void a3A(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.h;
	_core->de.i8.h = (btmp >> 1);

	_f_shift(_core, _core->de.i8.h, (btmp & 0x01));
}
/*	srl e
 *	.235
 */
static void a3B(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->de.i8.l;
	_core->de.i8.l = (btmp >> 1);

	_f_shift(_core, _core->de.i8.l, (btmp & 0x01));
}
/*	srl h
 *	.235
 */
static void a3C(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.h;
	_core->hl.i8.h = (btmp >> 1);

	_f_shift(_core, _core->hl.i8.h, (btmp & 0x01));
}
/*	srl l
 *	.235
 */
static void a3D(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->hl.i8.l;
	_core->hl.i8.l = (btmp >> 1);

	_f_shift(_core, _core->hl.i8.l, (btmp & 0x01));
}
/*	srl (ii+d)
 *	.235
 */
static void ia38(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, ii_addr);
	_cycle(_core, 1);
	val = (btmp >> 1);

	_f_shift(_core, val, (btmp & 0x01));
	_st_b(_core, ii_addr, val);
	*store = val;
}
/*	srl (hl)
 *	.235
 */
static void a3E(_core_o *_core)
{
	uint8 btmp, val;

	btmp = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 1);
	val = (btmp >> 1);

	_f_shift(_core, val, (btmp & 0x1));
	_st_b(_core, _core->hl.i16, val);
}
/*	srl a
 *	.235
 */
static void a3F(_core_o *_core)
{
	uint8 btmp;

	btmp = _core->af.i8.h;
	_core->af.i8.h = (btmp >> 1);

	_f_shift(_core, _core->af.i8.h, (btmp & 0x01));
}

INTERNAL void _bit(_core_o *_core, uint8 val, uint8 mask)
{
	uint8 res;
	uint8 f;

	res = val & mask;

	f = (res & (SFLAG | BITS35)) | HFLAG;

	if (!res)
		f |= ZFLAG | VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | f;
}

INTERNAL void _bit_addr(_core_o *_core, uint16 addr, uint8 mask)
{
	uint8 val, res;
	uint8 f;

	val = _ld_b(_core, addr);
	_cycle(_core, 1);
	
	res = val & mask;

	f = (res & SFLAG) | HFLAG | ((_core->addr >> 8) & BITS35);

	if (!res)
		f |= ZFLAG | VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | f;
}

/*	bit b, (ii+d)
 *	.246
 */
static void ia40(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x01);
}
static void ia48(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x02);
}
static void ia50(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x04);
}
static void ia58(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x08);
}
static void ia60(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x10);
}
static void ia68(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x20);
}
static void ia70(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x40);
}
static void ia78(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	_bit_addr(_core, ii_addr, 0x80);
}

/*	bit b, (hl)
 *	.244
 */
static void a46(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x01);
}
static void a4E(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x02);
}
static void a56(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x04);
}
static void a5E(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x08);
}
static void a66(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x10);
}
static void a6E(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x20);
}
static void a76(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x40);
}
static void a7E(_core_o *_core)
{
	_bit_addr(_core, _core->hl.i16, 0x80);
}

/*	bit b, b
 *	.242
 */
static void a40(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x01);
}
static void a48(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x02);
}
static void a50(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x04);
}
static void a58(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x08);
}
static void a60(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x10);
}
static void a68(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x20);
}
static void a70(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x40);
}
static void a78(_core_o *_core)
{
	_bit(_core, _core->bc.i8.h, 0x80);
}

/*	bit b, c
 *	.242
 */
static void a41(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x01);
}
static void a49(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x02);
}
static void a51(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x04);
}
static void a59(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x08);
}
static void a61(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x10);
}
static void a69(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x20);
}
static void a71(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x40);
}
static void a79(_core_o *_core)
{
	_bit(_core, _core->bc.i8.l, 0x80);
}

/*	bit b, d
 *	.242
 */
static void a42(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x01);
}
static void a4A(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x02);
}
static void a52(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x04);
}
static void a5A(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x08);
}
static void a62(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x10);
}
static void a6A(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x20);
}
static void a72(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x40);
}
static void a7A(_core_o *_core)
{
	_bit(_core, _core->de.i8.h, 0x80);
}

/*	bit b, e
 *	.242
 */
static void a43(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x01);
}
static void a4B(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x02);
}
static void a53(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x04);
}
static void a5B(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x08);
}
static void a63(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x10);
}
static void a6B(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x20);
}
static void a73(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x40);
}
static void a7B(_core_o *_core)
{
	_bit(_core, _core->de.i8.l, 0x80);
}

/*	bit b, h
 *	.242
 */
static void a44(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x01);
}
static void a4C(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x02);
}
static void a54(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x04);
}
static void a5C(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x08);
}
static void a64(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x10);
}
static void a6C(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x20);
}
static void a74(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x40);
}
static void a7C(_core_o *_core)
{
	_bit(_core, _core->hl.i8.h, 0x80);
}

/*	bit b, l
 *	.242
 */
static void a45(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x01);
}
static void a4D(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x02);
}
static void a55(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x04);
}
static void a5D(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x08);
}
static void a65(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x10);
}
static void a6D(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x20);
}
static void a75(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x40);
}
static void a7D(_core_o *_core)
{
	_bit(_core, _core->hl.i8.l, 0x80);
}

/*	bit b, a
 *	.242
 */
static void a47(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x01);
}
static void a4F(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x02);
}
static void a57(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x04);
}
static void a5F(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x08);
}
static void a67(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x10);
}
static void a6F(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x20);
}
static void a77(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x40);
}
static void a7F(_core_o *_core)
{
	_bit(_core, _core->af.i8.h, 0x80);
}

INTERNAL void _res(_core_o *_core, uint8 *val, uint8 mask)
{
	*val = *val & ~mask;
}

INTERNAL uint8 _res_addr(_core_o *_core, uint16 addr, uint8 mask)
{
	uint8 val;

	val = _ld_b(_core, addr);
	_cycle(_core, 1);

	val = val & ~mask;

	_st_b(_core, addr, val);

	return val;
}

/*	res b, (ii+d)
 *	.254
 */
static void ia80(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x01);
}
static void ia88(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x02);
}
static void ia90(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x04);
}
static void ia98(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x08);
}
static void iaA0(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x10);
}
static void iaA8(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x20);
}
static void iaB0(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x40);
}
static void iaB8(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _res_addr(_core, ii_addr, 0x80);
}

/*	res b, (hl)
 *	.254
 */
static void a86(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x01);
}
static void a8E(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x02);
}
static void a96(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x04);
}
static void a9E(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x08);
}
static void aA6(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x10);
}
static void aAE(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x20);
}
static void aB6(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x40);
}
static void aBE(_core_o *_core)
{
	_res_addr(_core, _core->hl.i16, 0x80);
}

/*	res b, b
 *	.254
 */
static void a80(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x01);
}
static void a88(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x02);
}
static void a90(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x04);
}
static void a98(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x08);
}
static void aA0(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x10);
}
static void aA8(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x20);
}
static void aB0(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x40);
}
static void aB8(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.h), 0x80);
}

/*	res b, c
 *	.254
 */
static void a81(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x01);
}
static void a89(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x02);
}
static void a91(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x04);
}
static void a99(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x08);
}
static void aA1(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x10);
}
static void aA9(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x20);
}
static void aB1(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x40);
}
static void aB9(_core_o *_core)
{
	_res(_core, &(_core->bc.i8.l), 0x80);
}

/*	res b, d
 *	.254
 */
static void a82(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x01);
}
static void a8A(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x02);
}
static void a92(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x04);
}
static void a9A(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x08);
}
static void aA2(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x10);
}
static void aAA(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x20);
}
static void aB2(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x40);
}
static void aBA(_core_o *_core)
{
	_res(_core, &(_core->de.i8.h), 0x80);
}

/*	res b, e
 *	.254
 */
static void a83(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x01);
}
static void a8B(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x02);
}
static void a93(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x04);
}
static void a9B(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x08);
}
static void aA3(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x10);
}
static void aAB(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x20);
}
static void aB3(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x40);
}
static void aBB(_core_o *_core)
{
	_res(_core, &(_core->de.i8.l), 0x80);
}

/*	res b, h
 *	.254
 */
static void a84(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x01);
}
static void a8C(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x02);
}
static void a94(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x04);
}
static void a9C(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x08);
}
static void aA4(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x10);
}
static void aAC(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x20);
}
static void aB4(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x40);
}
static void aBC(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.h), 0x80);
}

/*	res b, l
 *	.254
 */
static void a85(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x01);
}
static void a8D(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x02);
}
static void a95(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x04);
}
static void a9D(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x08);
}
static void aA5(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x10);
}
static void aAD(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x20);
}
static void aB5(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x40);
}
static void aBD(_core_o *_core)
{
	_res(_core, &(_core->hl.i8.l), 0x80);
}

/*	res b, a
 *	.254
 */
static void a87(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x01);
}
static void a8F(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x02);
}
static void a97(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x04);
}
static void a9F(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x08);
}
static void aA7(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x10);
}
static void aAF(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x20);
}
static void aB7(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x40);
}
static void aBF(_core_o *_core)
{
	_res(_core, &(_core->af.i8.h), 0x80);
}

INTERNAL void _set(_core_o *_core, uint8 *val, uint8 mask)
{
	*val = *val | mask;
}

INTERNAL uint8 _set_addr(_core_o *_core, uint16 addr, uint8 mask)
{
	uint8 val;

	val = _ld_b(_core, addr);
	_cycle(_core, 1);

	val = val | mask;

	_st_b(_core, addr, val);

	return val;
}

/*	set b, (ii+d)
 *	.254
 */
static void iaC0(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x01);
}
static void iaC8(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x02);
}
static void iaD0(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x04);
}
static void iaD8(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x08);
}
static void iaE0(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x10);
}
static void iaE8(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x20);
}
static void iaF0(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x40);
}
static void iaF8(_core_o *_core, uint16 ii_addr, uint8 *store)
{
	*store = _set_addr(_core, ii_addr, 0x80);
}

/*	set b, (hl)
 *	.254
 */
static void aC6(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x01);
}
static void aCE(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x02);
}
static void aD6(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x04);
}
static void aDE(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x08);
}
static void aE6(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x10);
}
static void aEE(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x20);
}
static void aF6(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x40);
}
static void aFE(_core_o *_core)
{
	_set_addr(_core, _core->hl.i16, 0x80);
}

/*	set b, b
 *	.254
 */
static void aC0(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x01);
}
static void aC8(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x02);
}
static void aD0(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x04);
}
static void aD8(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x08);
}
static void aE0(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x10);
}
static void aE8(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x20);
}
static void aF0(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x40);
}
static void aF8(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.h), 0x80);
}

/*	set b, c
 *	.254
 */
static void aC1(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x01);
}
static void aC9(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x02);
}
static void aD1(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x04);
}
static void aD9(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x08);
}
static void aE1(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x10);
}
static void aE9(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x20);
}
static void aF1(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x40);
}
static void aF9(_core_o *_core)
{
	_set(_core, &(_core->bc.i8.l), 0x80);
}

/*	set b, d
 *	.254
 */
static void aC2(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x01);
}
static void aCA(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x02);
}
static void aD2(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x04);
}
static void aDA(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x08);
}
static void aE2(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x10);
}
static void aEA(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x20);
}
static void aF2(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x40);
}
static void aFA(_core_o *_core)
{
	_set(_core, &(_core->de.i8.h), 0x80);
}

/*	set b, e
 *	.254
 */
static void aC3(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x01);
}
static void aCB(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x02);
}
static void aD3(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x04);
}
static void aDB(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x08);
}
static void aE3(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x10);
}
static void aEB(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x20);
}
static void aF3(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x40);
}
static void aFB(_core_o *_core)
{
	_set(_core, &(_core->de.i8.l), 0x80);
}

/*	set b, h
 *	.254
 */
static void aC4(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x01);
}
static void aCC(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x02);
}
static void aD4(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x04);
}
static void aDC(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x08);
}
static void aE4(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x10);
}
static void aEC(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x20);
}
static void aF4(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x40);
}
static void aFC(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.h), 0x80);
}

/*	set b, l
 *	.254
 */
static void aC5(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x01);
}
static void aCD(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x02);
}
static void aD5(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x04);
}
static void aDD(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x08);
}
static void aE5(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x10);
}
static void aED(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x20);
}
static void aF5(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x40);
}
static void aFD(_core_o *_core)
{
	_set(_core, &(_core->hl.i8.l), 0x80);
}

/*	set b, a
 *	.254
 */
static void aC7(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x01);
}
static void aCF(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x02);
}
static void aD7(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x04);
}
static void aDF(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x08);
}
static void aE7(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x10);
}
static void aEF(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x20);
}
static void aF7(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x40);
}
static void aFF(_core_o *_core)
{
	_set(_core, &(_core->af.i8.h), 0x80);
}

INTERNAL uint8 _in_r_C(_core_o *_core)
{
	uint8 f, val;

	val = _in(_core, _core->bc.i16);

	f = (val & (SFLAG | BITS35));

	if (!val)
		f |= ZFLAG;

	if (_parity(val))
		f |= VFLAG;

	_core->af.i8.l = f;

	return val;
}

/*	in b, (C)
 *	.288
 */
static void x40(_core_o *_core)
{
	_core->bc.i8.h = _in_r_C(_core);
}
/*	in c, (C)
 *	.288
 */
static void x48(_core_o *_core)
{
	_core->bc.i8.l = _in_r_C(_core);
}
/*	in d, (C)
 *	.288
 */
static void x50(_core_o *_core)
{
	_core->de.i8.h = _in_r_C(_core);
}
/*	in e, (C)
 *	.288
 */
static void x58(_core_o *_core)
{
	_core->de.i8.l = _in_r_C(_core);
}
/*	in h, (C)
 *	.288
 */
static void x60(_core_o *_core)
{
	_core->hl.i8.h = _in_r_C(_core);
}
/*	in l, (C)
 *	.288
 */
static void x68(_core_o *_core)
{
	_core->hl.i8.l = _in_r_C(_core);
}
/*	in f, (C)
 *	.288
 */
static void x70(_core_o *_core)
{
	_in_r_C(_core);
}
/*	in a, (C)
 *	.288
 */
static void x78(_core_o *_core)
{
	_core->af.i8.h = _in_r_C(_core);
}

/*	out (C), b
 *	.298
 */
static void x41(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->bc.i8.h);
}
/*	out (C), c
 *	.298
 */
static void x49(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->bc.i8.l);
}
/*	out (C), d
 *	.298
 */
static void x51(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->de.i8.h);
}
/*	out (C), e
 *	.298
 */
static void x59(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->de.i8.l);
}
/*	out (C), h
 *	.298
 */
static void x61(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->hl.i8.h);
}
/*	out (C), l
 *	.298
 */
static void x69(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->hl.i8.l);
}
/*	out (C), 0
 *	.298
 */
static void x71(_core_o *_core)
{
	_out(_core, _core->bc.i16, 0);
}
/*	out (C), a
 *	.298
 */
static void x79(_core_o *_core)
{
	_out(_core, _core->bc.i16, _core->af.i8.h);
}

/*	ld (nn), bc
 *	.128
 */
static void x43(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_st_w(_core, addr, _core->bc.i16);
}
/*	ld (nn), de
 *	.128
 */
static void x53(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_st_w(_core, addr, _core->de.i16);
}
/*	ld (nn), hl
 *	.128
 */
static void x63(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_st_w(_core, addr, _core->hl.i16);
}
/*	ld (nn), sp
 *	.128
 */
static void x73(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_st_w(_core, addr, _core->sp);
}

/*	ld bc, (nn)
 *	.124
 */
static void x4B(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_core->bc.i16 = _ld_w(_core, addr);
}
/*	ld de, (nn)
 *	.124
 */
static void x5B(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_core->de.i16 = _ld_w(_core, addr);
}
/*	ld hl, (nn)
 *	.124
 */
static void x6B(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_core->hl.i16 = _ld_w(_core, addr);
}
/*	ld sp, (nn)
 *	.124
 */
static void x7B(_core_o *_core)
{
	uint16 addr;

	addr = _imm_w(_core);

	_core->sp = _ld_w(_core, addr);
}

/*	neg
 *	.187
 */
static void x44(_core_o *_core)
{
	uint8 a, val, f;
	uint16 res;

	a = 0;
	val = _core->af.i8.h;

	res = -((uint16)val);

	f = (res & (SFLAG | BITS35)) | ((res >> 8) & CFLAG);

	if (!(res & 0xff))
		f |= ZFLAG;

	f |= ((val ^ res) & HFLAG);

	if (val == 0x80)
		f |= VFLAG;

	_core->af.i8.l = f | NFLAG;
	_core->af.i8.h = res;
}

/*	im 0
 *	.194
 */
static void x46(_core_o *_core)
{
	_core->im = 0;
}

/*	im 1
 *	.195
 */
static void x56(_core_o *_core)
{
	_core->im = 1;
}

/*	im 2
 *	.196
 */
static void x5E(_core_o *_core)
{
	_core->im = 2;
}

/*	ld i, a
 *	.118
 */
static void x47(_core_o *_core)
{
	_core->i = _core->af.i8.h;

	_cycle(_core, 1);
}

/*	ld a, i
 *	.116
 */
static void x57(_core_o *_core)
{
	uint8 f;
	uint8 res;

	res = _core->i;
	_cycle(_core, 1);

	f = (res & (SFLAG | BITS35));

	if (!_core->iff1)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | (f & cFLAG);
	_core->af.i8.h = res;
}

/*	ld r, a
 *	.119
 */
static void x4F(_core_o *_core)
{
	_core->r = ST_R(_core->af.i8.h);

	_cycle(_core, 1);
}

/*	ld a, r
 *	.117
 */
static void x5F(_core_o *_core)
{
	uint8 f;
	uint8 res;

	res = LD_R(_core);

	_cycle(_core, 1);

	f = (res & (SFLAG | BITS35));

	if (!_core->iff1)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | (f & cFLAG);
	_core->af.i8.h = res;
}

/*	retn
 *	.283
 */
static void x45(_core_o *_core)
{
	_core->iff1 = _core->iff2;

	_ret(_core);
}

/*	reti
 *	.281
 */
static void x4D(_core_o *_core)
{
	_core->iff1 = _core->iff2;

	_ret(_core);
}

/*	rrd
 *	.240
 */
static void x67(_core_o *_core)
{
	uint8 f;
	uint8 r, n;

	r = _core->af.i8.h;
	n = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 4);
	_st_b(_core, _core->hl.i16, (r << 4) | (n >> 4));

	r = (r & 0xf0) | (n & 0x0f);

	f = (r & (SFLAG | BITS35));

	if (!r)
		f |= ZFLAG;

	if (_parity(r))
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | f;
	_core->af.i8.h = r;
}

/*	rld
 *	.238
 */
static void x6F(_core_o *_core)
{
	uint8 f;
	uint8 r, n;

	r = _core->af.i8.h;
	n = _ld_b(_core, _core->hl.i16);
	_cycle(_core, 4);
	_st_b(_core, _core->hl.i16, (n << 4) | (r & 0x0f));

	r = (r & 0xf0) | (n >> 4);

	f = (r & (SFLAG | BITS35));

	if (!r)
		f |= ZFLAG;

	if (_parity(r))
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | f;
	_core->af.i8.h = r;
}

INTERNAL void _sbc_hl_d(_core_o *_core, uint16 val)
{
	uint32 res;
	uint16 hl = _core->hl.i16;
	uint8 f;

	res = (uint32)hl - (uint32)val - (_core->af.i8.l & CFLAG);

	f = ((res >> 8) & (SFLAG | BITS35)) | ((res >> 16) & CFLAG);

	if (!(res & 0xFFFF))
		f |= ZFLAG;

	if ((hl ^ val ^ res) & 0x1000)
		f |= HFLAG;

	if ((hl ^ val) & (hl ^ res) & 0x8000)
		f |= VFLAG;

	_core->af.i8.l = f | NFLAG;
	_core->addr = _core->hl.i16 = res;

	_cycle(_core, 7);
}

/*	sbc hl, bc
 *	.199
 */
static void x42(_core_o *_core)
{
	_sbc_hl_d(_core, _core->bc.i16);
}
/*	sbc hl, de
 *	.199
 */
static void x52(_core_o *_core)
{
	_sbc_hl_d(_core, _core->de.i16);
}
/*	sbc hl, hl
 *	.199
 */
static void x62(_core_o *_core)
{
	_sbc_hl_d(_core, _core->hl.i16);
}
/*	sbc hl, sp
 *	.199
 */
static void x72(_core_o *_core)
{
	_sbc_hl_d(_core, _core->sp);
}

INTERNAL void _adc_hl_d(_core_o *_core, uint16 val)
{
	uint32 res;
	uint16 hl = _core->hl.i16;
	uint8 f;

	res = (uint32)hl + (uint32)val + (_core->af.i8.l & CFLAG);

	f = ((res >> 8) & (SFLAG | BITS35)) | ((res >> 16) & CFLAG);

	if (!(res & 0xFFFF))
		f |= ZFLAG;

	if ((hl ^ val ^ res) & 0x1000)
		f |= HFLAG;

	if (~(hl ^ val) & (hl ^ res) & 0x8000)
		f |= VFLAG;

	_core->af.i8.l = f;
	_core->addr = _core->hl.i16 = res;

	_cycle(_core, 7);
}

/*	adc hl, bc
 *	.198
 */
static void x4A(_core_o *_core)
{
	_adc_hl_d(_core, _core->bc.i16);
}
/*	adc hl, de
 *	.198
 */
static void x5A(_core_o *_core)
{
	_adc_hl_d(_core, _core->de.i16);
}
/*	adc hl, hl
 *	.198
 */
static void x6A(_core_o *_core)
{
	_adc_hl_d(_core, _core->hl.i16);
}
/*	adc hl, sp
 *	.198
 */
static void x7A(_core_o *_core)
{
	_adc_hl_d(_core, _core->sp);
}


INTERNAL void _ldi(_core_o *_core)
{
	uint8 f, data;

	data = _ld_b(_core, _core->hl.i16);
	_st_b(_core, _core->de.i16, data);

	_cycle(_core, 2);

	data += _core->af.i8.h;
	f = (data & XFLAG);
	if (data & 0x02)
		f |= YFLAG;

	_core->hl.i16++;
	_core->de.i16++;
	_core->bc.i16--;

	if (_core->bc.i16)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | CFLAG)) | f;
}

/*	ldi
 *	.146
 */
static void xA0(_core_o *_core)
{
	_ldi(_core);
}
/*	ldir
 *	.147
 */
static void xB0(_core_o *_core)
{
	_ldi(_core);

	if(!_core->bc.i16)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _ldd(_core_o *_core)
{
	uint8 f, data;

	data = _ld_b(_core, _core->hl.i16);
	_st_b(_core, _core->de.i16, data);

	_cycle(_core, 2);

	data += _core->af.i8.h;
	f = (data & XFLAG);
	if (data & 0x02)
		f |= YFLAG;

	_core->hl.i16--;
	_core->de.i16--;
	_core->bc.i16--;

	if (_core->bc.i16)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & (SFLAG | ZFLAG | CFLAG)) | f;
}

/*	ldd
 *	.149
 */
static void xA8(_core_o *_core)
{
	_ldd(_core);
}
/*	lddr
 *	.150
 */
static void xB8(_core_o *_core)
{
	_ldd(_core);

	if(!_core->bc.i16)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _cpi(_core_o *_core)
{
	uint8 a, f;
	uint8 val;
	uint8 res;

	a = _core->af.i8.h;
  	val = _ld_b(_core, _core->hl.i16);

	res = a - val;
	_core->hl.i16++;
	_core->bc.i16--;

	f = (res & SFLAG) | NFLAG;

	if (!res)
		f |= ZFLAG;

	f |= ((a ^ val ^ res) & HFLAG);

	if (f & HFLAG)
		res--;

	if (res & 0x02)
		f |= YFLAG;
	f |= (res & XFLAG);

	if(_core->bc.i16)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | (f & cFLAG);
}

/*	cpi
 *	.152
 */
static void xA1(_core_o *_core)
{
	_cpi(_core);
}
/*	cpir
 *	.153
 */
static void xB1(_core_o *_core)
{
	_cpi(_core);

	if (_core->af.i8.l & ZFLAG)
		return;

	if (!_core->bc.i16)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _cpd(_core_o *_core)
{
	uint8 a, f;
	uint8 val;
	uint8 res;

	a = _core->af.i8.h;
  	val = _ld_b(_core, _core->hl.i16);

	res = a - val;
	_core->hl.i16--;
	_core->bc.i16--;

	f = (res & SFLAG) | NFLAG;

	if (!res)
		f |= ZFLAG;

	f |= ((a ^ val ^ res) & HFLAG);

	if (f & HFLAG)
		res--;

	if (res & 0x02)
		f |= YFLAG;
	f |= (res & XFLAG);

	if(_core->bc.i16)
		f |= VFLAG;

	_core->af.i8.l = (_core->af.i8.l & CFLAG) | (f & cFLAG);
}

/*	cpd
 *	.155
 */
static void xA9(_core_o *_core)
{
	_cpd(_core);
}
/*	cpdr
 *	.156
 */
static void xB9(_core_o *_core)
{
	_cpd(_core);

	if (_core->af.i8.l & ZFLAG)
		return;

	if (!_core->bc.i16)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _ini(_core_o *_core)
{
	uint8 data, f;

	data = _in(_core, _core->bc.i16);
	_st_b(_core, _core->hl.i16, data);
	_core->hl.i16++;
	_core->bc.i8.h--;

	f = NFLAG;

	if (!_core->bc.i8.h)
		f |= ZFLAG;
}

static void xA2(_core_o *_core)
{
	_ini(_core);
}
static void xB2(_core_o *_core)
{
	_ini(_core);

	if (!_core->bc.i8.h)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _ind(_core_o *_core)
{
	uint8 data, f;

	data = _in(_core, _core->bc.i16);
	_st_b(_core, _core->hl.i16, data);
	_core->hl.i16--;
	_core->bc.i8.h--;

	f = NFLAG;

	if (!_core->bc.i8.h)
		f |= ZFLAG;
}

static void xAA(_core_o *_core)
{
	_ind(_core);
}
static void xBA(_core_o *_core)
{
	_ind(_core);

	if (!_core->bc.i8.h)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _outi(_core_o *_core)
{
	uint8 data, f;

	data = _ld_b(_core, _core->hl.i16);
	_out(_core, _core->bc.i16, data);

	_core->hl.i16++;
	_core->bc.i8.h--;

	f = NFLAG;

	if (!_core->bc.i8.h)
		f |= ZFLAG;
}

static void xA3(_core_o *_core)
{
	_outi(_core);
}
static void xB3(_core_o *_core)
{
	_outi(_core);

	if (!_core->bc.i8.h)
		return;

	_jr(_core, 0xfffe);
}

INTERNAL void _outd(_core_o *_core)
{
	uint8 data, f;

	data = _ld_b(_core, _core->hl.i16);
	_out(_core, _core->bc.i16, data);

	_core->hl.i16--;
	_core->bc.i8.h--;

	f = NFLAG;

	if (!_core->bc.i8.h)
		f |= ZFLAG;
}

static void xAB(_core_o *_core)
{
	_outd(_core);
}
static void xBB(_core_o *_core)
{
	_outd(_core);

	if (!_core->bc.i8.h)
		return;

	_jr(_core, 0xfffe);
}

static z80_opcode z80_opc_main_opcodes[256] = {
 m00, m01, m02, m03, m04, m05, m06, m07, m08, m09, m0A, m0B, m0C, m0D, m0E, m0F,
 m10, m11, m12, m13, m14, m15, m16, m17, m18, m19, m1A, m1B, m1C, m1D, m1E, m1F,
 m20, m21, m22, m23, m24, m25, m26, m27, m28, m29, m2A, m2B, m2C, m2D, m2E, m2F,
 m30, m31, m32, m33, m34, m35, m36, m37, m38, m39, m3A, m3B, m3C, m3D, m3E, m3F,
 m40, m41, m42, m43, m44, m45, m46, m47, m48, m49, m4A, m4B, m4C, m4D, m4E, m4F,
 m50, m51, m52, m53, m54, m55, m56, m57, m58, m59, m5A, m5B, m5C, m5D, m5E, m5F,
 m60, m61, m62, m63, m64, m65, m66, m67, m68, m69, m6A, m6B, m6C, m6D, m6E, m6F,
 m70, m71, m72, m73, m74, m75, m76, m77, m78, m79, m7A, m7B, m7C, m7D, m7E, m7F,
 m80, m81, m82, m83, m84, m85, m86, m87, m88, m89, m8A, m8B, m8C, m8D, m8E, m8F,
 m90, m91, m92, m93, m94, m95, m96, m97, m98, m99, m9A, m9B, m9C, m9D, m9E, m9F,
 mA0, mA1, mA2, mA3, mA4, mA5, mA6, mA7, mA8, mA9, mAA, mAB, mAC, mAD, mAE, mAF,
 mB0, mB1, mB2, mB3, mB4, mB5, mB6, mB7, mB8, mB9, mBA, mBB, mBC, mBD, mBE, mBF,
 mC0, mC1, mC2, mC3, mC4, mC5, mC6, mC7, mC8, mC9, mCA, mCB, mCC, mCD, mCE, mCF,
 mD0, mD1, mD2, mD3, mD4, mD5, mD6, mD7, mD8, mD9, mDA, mDB, mDC, mDD, mDE, mDF,
 mE0, mE1, mE2, mE3, mE4, mE5, mE6, mE7, mE8, mE9, mEA, mEB, mEC, mED, mEE, mEF,
 mF0, mF1, mF2, mF3, mF4, mF5, mF6, mF7, mF8, mF9, mFA, mFB, mFC, mFD, mFE, mFF,
};

static z80_opcode z80_opc_extended_opcodes[256] = {
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x40, x41, x42, x43, x44, x45, x46, x47, x48, x49, x4A, x4B, x44, x4D, x46, x4F,
 x50, x51, x52, x53, x44, x45, x56, x57, x58, x59, x5A, x5B, x44, x4D, x5E, x5F,
 x60, x61, x62, x63, x44, x45, x46, x67, x68, x69, x6A, x6B, x44, x4D, x46, x6F,
 x70, x71, x72, x73, x44, x45, x56, x00, x78, x79, x7A, x7B, x44, x4D, x5E, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 xA0, xA1, xA2, xA3, x00, x00, x00, x00, xA8, xA9, xAA, xAB, x00, x00, x00, x00,
 xB0, xB1, xB2, xB3, x00, x00, x00, x00, xB8, xB9, xBA, xBB, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
 x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00, x00,
};

static z80_opcode z80_opc_bit_opcodes[256] = {
 a00, a01, a02, a03, a04, a05, a06, a07, a08, a09, a0A, a0B, a0C, a0D, a0E, a0F,
 a10, a11, a12, a13, a14, a15, a16, a17, a18, a19, a1A, a1B, a1C, a1D, a1E, a1F,
 a20, a21, a22, a23, a24, a25, a26, a27, a28, a29, a2A, a2B, a2C, a2D, a2E, a2F,
 a30, a31, a32, a33, a34, a35, a36, a37, a38, a39, a3A, a3B, a3C, a3D, a3E, a3F,
 a40, a41, a42, a43, a44, a45, a46, a47, a48, a49, a4A, a4B, a4C, a4D, a4E, a4F,
 a50, a51, a52, a53, a54, a55, a56, a57, a58, a59, a5A, a5B, a5C, a5D, a5E, a5F,
 a60, a61, a62, a63, a64, a65, a66, a67, a68, a69, a6A, a6B, a6C, a6D, a6E, a6F,
 a70, a71, a72, a73, a74, a75, a76, a77, a78, a79, a7A, a7B, a7C, a7D, a7E, a7F,
 a80, a81, a82, a83, a84, a85, a86, a87, a88, a89, a8A, a8B, a8C, a8D, a8E, a8F,
 a90, a91, a92, a93, a94, a95, a96, a97, a98, a99, a9A, a9B, a9C, a9D, a9E, a9F,
 aA0, aA1, aA2, aA3, aA4, aA5, aA6, aA7, aA8, aA9, aAA, aAB, aAC, aAD, aAE, aAF,
 aB0, aB1, aB2, aB3, aB4, aB5, aB6, aB7, aB8, aB9, aBA, aBB, aBC, aBD, aBE, aBF,
 aC0, aC1, aC2, aC3, aC4, aC5, aC6, aC7, aC8, aC9, aCA, aCB, aCC, aCD, aCE, aCF,
 aD0, aD1, aD2, aD3, aD4, aD5, aD6, aD7, aD8, aD9, aDA, aDB, aDC, aDD, aDE, aDF,
 aE0, aE1, aE2, aE3, aE4, aE5, aE6, aE7, aE8, aE9, aEA, aEB, aEC, aED, aEE, aEF,
 aF0, aF1, aF2, aF3, aF4, aF5, aF6, aF7, aF8, aF9, aFA, aFB, aFC, aFD, aFE, aFF,
};

static z80_opcode z80_opc_indexed_opcodes[256] = {
 m00, m01, m02, m03, m04, m05, m06, m07, m08, m09, m0A, m0B, m0C, m0D, m0E, m0F,
 m10, m11, m12, m13, m14, m15, m16, m17, m18, m19, m1A, m1B, m1C, m1D, m1E, m1F,
 m20, m21, m22, m23, m24, m25, m26, m27, m28, m29, m2A, m2B, m2C, m2D, m2E, m2F,
 m30, m31, m32, m33, i34, i35, i36, m37, m38, m39, m3A, m3B, m3C, m3D, m3E, m3F,
 m40, m41, m42, m43, m44, m45, i46, m47, m48, m49, m4A, m4B, m4C, m4D, i4E, m4F,
 m50, m51, m52, m53, m54, m55, i56, m57, m58, m59, m5A, m5B, m5C, m5D, i5E, m5F,
 m60, m61, m62, m63, m64, m65, i66, m67, m68, m69, m6A, m6B, m6C, m6D, i6E, m6F,
 i70, i71, i72, i73, i74, i75, m76, i77, m78, m79, m7A, m7B, m7C, m7D, i7E, m7F,
 m80, m81, m82, m83, m84, m85, i86, m87, m88, m89, m8A, m8B, m8C, m8D, i8E, m8F,
 m90, m91, m92, m93, m94, m95, i96, m97, m98, m99, m9A, m9B, m9C, m9D, i9E, m9F,
 mA0, mA1, mA2, mA3, mA4, mA5, iA6, mA7, mA8, mA9, mAA, mAB, mAC, mAD, iAE, mAF,
 mB0, mB1, mB2, mB3, mB4, mB5, iB6, mB7, mB8, mB9, mBA, mBB, mBC, mBD, iBE, mBF,
 mC0, mC1, mC2, mC3, mC4, mC5, mC6, mC7, mC8, mC9, mCA, iCB, mCC, mCD, mCE, mCF,
 mD0, mD1, mD2, mD3, mD4, mD5, mD6, mD7, mD8, mD9, mDA, mDB, mDC, mDD, mDE, mDF,
 mE0, mE1, mE2, mE3, mE4, mE5, mE6, mE7, mE8, mE9, mEA, mEB, mEC, mED, mEE, mEF,
 mF0, mF1, mF2, mF3, mF4, mF5, mF6, mF7, mF8, mF9, mFA, mFB, mFC, mFD, mFE, mFF,
};

static z80_ia_opcode z80_opc_index_bit_opcodes[32] = {
	ia00, ia08,
	ia10, ia18,
	ia20, ia28,
	ia30, ia38,
	ia40, ia48,
	ia50, ia58,
	ia60, ia68,
	ia70, ia78,
	ia80, ia88,
	ia90, ia98,
	iaA0, iaA8,
	iaB0, iaB8,
	iaC0, iaC8,
	iaD0, iaD8,
	iaE0, iaE8,
	iaF0, iaF8
};

static void iCB(_core_o *_core)
{
	uint16 addr;
	uint8 *store;
	uint8 zero;

	_core->r += 2;

	addr = _addr_ii(_core);

	_fetch(_core);

	store = &zero;

	switch (_core->current_opc & 0x7) {
		case 0:	store = &(_core->bc.i8.h); break;
		case 1:	store = &(_core->bc.i8.l); break;
		case 2:	store = &(_core->de.i8.h); break;
		case 3:	store = &(_core->de.i8.l); break;
		case 4:	store = &(_core->hl.i8.h); break;
		case 5:	store = &(_core->hl.i8.l); break;
		case 7:	store = &(_core->af.i8.h); break;
	}

	z80_opc_index_bit_opcodes[_core->current_opc >> 3](_core, addr, store);
}

static void mCB(_core_o *_core)
{
	_core->r += 2;

	_fetch(_core);

#ifdef DEBUG_USE
	if (!_used_bit[_core->current_opc]) {
		_used_bit[_core->current_opc] = 1;
		fprintf(stderr, "[Z80/DEC] %04X: CB%02X\n", _core->pc-1, _core->current_opc);
	}
#endif

	z80_opc_bit_opcodes[_core->current_opc](_core);
}

static void mED(_core_o *_core)
{
	_core->r += 2;

	_fetch(_core);

#ifdef DEBUG_USE
	if (!_used_extended[_core->current_opc]) {
		_used_extended[_core->current_opc] = 1;
		fprintf(stderr, "[Z80/DEC] %04X: ED%02X\n", _core->pc-1, _core->current_opc);
	}
#endif

	z80_opc_extended_opcodes[_core->current_opc](_core);
}

static void mDD(_core_o *_core)
{
	_core->r += 2;
	_core->ii = &(_core->ix);

	_fetch(_core);

#ifdef DEBUG_USE
	if (!_used_indexed[_core->current_opc]) {
		_used_indexed[_core->current_opc] = 1;
		fprintf(stderr, "[Z80/DEC] %04X: II%02X\n", _core->pc-1, _core->current_opc);
	}
#endif

	z80_opc_indexed_opcodes[_core->current_opc](_core);
}

static void mFD(_core_o *_core)
{
	_core->r += 2;
	_core->ii = &(_core->iy);

	_fetch(_core);

#ifdef DEBUG_USE
	if (!_used_indexed[_core->current_opc]) {
		_used_indexed[_core->current_opc] = 1;
		fprintf(stderr, "[Z80/DEC] %04X: II%02X\n", _core->pc-1, _core->current_opc);
	}
#endif

	z80_opc_indexed_opcodes[_core->current_opc](_core);
}

INTERNAL void _step(_core_o *_core)
{
	_core->r += 2;

	if (_core->halted) {
		_cycle(_core, 4);
		return;
	}

	_core->ii = &(_core->hl);
	_fetch(_core);

#ifdef DEBUG
	fprintf(stderr, "[Z80/DEC] %04X: %02X\n", _core->pc-1, _core->current_opc);
#endif

#ifdef DEBUG_USE
	if (!_used_main[_core->current_opc]) {
		_used_main[_core->current_opc] = 1;
		fprintf(stderr, "[Z80/DEC] %04X: %02X\n", _core->pc-1, _core->current_opc);
	}
#endif

	z80_opc_main_opcodes[_core->current_opc](_core);

	if (_core->active != 0) {
		_core->active--;
	} else if (_core->interrupted) {
		if (!_core->low_power)
			core_interrupt((core_o)_core, 0);

		_core->interrupted--;
	}
}

int core_execute(core_o core)
{
	_core_o *_core = (_core_o *)core;

	while (!_core->terminate) {
		_step(_core);

		if (_core->halted) {
			_core->r += ((_core->cycles >> 1) & 0xfe);

			_core->cycles = (_core->cycles & 0x3) - 4;
		}

		if (_core->cycles <= 0) {
			_core->cycles += _core->timings->cpint;

			if (!_core->low_power)
				core_interrupt(core, 0);

			_core->timings->callback(_core->timings);
		}
	}

	return 0;
}

void core_terminate(core_o core)
{
	_core_o *_core = (_core_o *)core;

	_core->terminate = 1;
}

int core_cycle(core_o core)
{
	_core_o *_core = (_core_o *)core;
	int cycles = _core->cycles;

	_step(_core);

	return (cycles - _core->cycles);
}

void core_interrupt(core_o core, unsigned int data)
{
	_core_o *_core = (_core_o *)core;
	uint16 addr;

	if (!_core->iff1)
		return;

	_core->iff1 = 0;

	if (_core->active) {
		_core->interrupted++;
		return;
	}

	if (_core->halted) {
		_core->pc++;
		_core->halted = 0;
	}

	if (_core->im == 2) {
		addr = (_core->i << 8);
		_call(_core, _ld_w(core, addr));
	} else if (_core->im == 1) {
		_call(_core, 0x38);			// rst 38h
	} else {
		_core->current_opc = data & 0xff;

		z80_opc_main_opcodes[_core->current_opc](_core);
	}
}
 
void core_call(core_o core, unsigned int addr)
{
	_core_o *_core = (_core_o *)core;

	_call(_core, addr & 0xffff);
}

void core_ret(core_o core)
{
	_core_o *_core = (_core_o *)core;

	_ret(_core);
}

void core_report(core_o core, FILE *out)
{
	_core_o *_core = (_core_o *)core;

	fprintf(out, "af : %04X\tbc : %04X\tde : %04X\thl : %04X\n",
		_core->af.i16, _core->bc.i16, _core->de.i16, _core->hl.i16);

	fprintf(out, "af': %04X\tbc': %04X\tde': %04X\thl': %04X\n",
		_core->af_, _core->bc_, _core->de_, _core->hl_);

	fprintf(out, "ix : %04X\tiy : %04X\tsp : %04X\tip : %04X\n",
		_core->ix.i16, _core->iy.i16, _core->sp, _core->pc);

	fprintf(out, "i  :   %02X\tr  :   %02X\tim :    %i\t",
		_core->i, LD_R(_core), _core->im);

	fprintf(out, "%c", (_core->af.i8.l & SFLAG) ? 'S' : 's');
	fprintf(out, "%c", (_core->af.i8.l & ZFLAG) ? 'Z' : 'z');
	fprintf(out, "%c", (_core->af.i8.l & YFLAG) ? 'Y' : 'y');
	fprintf(out, "%c", (_core->af.i8.l & HFLAG) ? 'H' : 'h');
	fprintf(out, "%c", (_core->af.i8.l & XFLAG) ? 'X' : 'x');
	fprintf(out, "%c", (_core->af.i8.l & VFLAG) ? 'V' : 'v');
	fprintf(out, "%c", (_core->af.i8.l & NFLAG) ? 'N' : 'n');
	fprintf(out, "%c\n", (_core->af.i8.l & CFLAG) ? 'C' : 'c');

	fprintf(out, "halted: %i\tactive: %i\tinterrupted: %i\n",
		_core->halted, _core->active, _core->interrupted);

	core_unassemble(core, _core->pc, out);
}

void core_alter(core_o core, const char *regname, unsigned int value)
{
	_core_o *_core = (_core_o *)core;

	if (!strcmp(regname, "a")) _core->af.i8.h = value;
	else if (!strcmp(regname, "f")) _core->af.i8.l = value;
	else if (!strcmp(regname, "b")) _core->bc.i8.h = value;
	else if (!strcmp(regname, "c")) _core->bc.i8.l = value;
	else if (!strcmp(regname, "d")) _core->de.i8.h = value;
	else if (!strcmp(regname, "e")) _core->de.i8.l = value;
	else if (!strcmp(regname, "h")) _core->hl.i8.h = value;
	else if (!strcmp(regname, "l")) _core->hl.i8.l = value;
	else if (!strcmp(regname, "af")) _core->af.i16 = value;
	else if (!strcmp(regname, "bc")) _core->bc.i16 = value;
	else if (!strcmp(regname, "de")) _core->de.i16 = value;
	else if (!strcmp(regname, "hl")) _core->hl.i16 = value;
	else if (!strcmp(regname, "af'")) _core->af_ = value;
	else if (!strcmp(regname, "bc'")) _core->bc_ = value;
	else if (!strcmp(regname, "de'")) _core->de_ = value;
	else if (!strcmp(regname, "hl'")) _core->hl_ = value;
	else if (!strcmp(regname, "ix")) _core->ix.i16 = value;
	else if (!strcmp(regname, "iy")) _core->iy.i16 = value;
	else if (!strcmp(regname, "sp")) _core->sp = value;
	else if (!strcmp(regname, "pc")) _core->pc = value;
	else if (!strcmp(regname, "i")) _core->i = value;
	else if (!strcmp(regname, "r")) _core->r = value;
}
