#include <SDL.h>
#include <stdio.h>
#include <stdlib.h>

#define VPCALC_SECTION	"GFX"
#define VPCALC_SUBSECT	"SDL"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "sdl.h"

//#undef OS_OSX

int sdl_initialized = 0;

typedef struct _sdl_o {
	SDL_Surface *image, *screen;

	union {
		uint32 *i32;
		uint16 *i16;
		uint8 *i8;
	} pixels;

	uint pitch, depth;
} _sdl_o;

sdl_o sdl_init(uint width, uint height, uint depth)
{
	_sdl_o *new;
	int rmask, gmask, bmask, amask;

#if SDL_BTYEORDER == SDL_BIG_ENDIAN
	if (depth == 32) {
		rmask = 0xff000000;
		gmask = 0x00ff0000;
		bmask = 0x0000ff00;
		amask = 0x00000000;
	} else if (depth == 16) {
		rmask = 0xf800;
		gmask = 0x07e0;
		bmask = 0x001f;
		amask = 0x0000;
	}
#else
	if (depth == 32) {
		rmask = 0x000000FF;
		gmask = 0x0000FF00;
		bmask = 0x00FF0000;
		amask = 0x00000000;
	} else if (depth == 16) {
		rmask = 0x001f;
		gmask = 0x07e0;
		bmask = 0xf800;
		amask = 0x0000;
	}
#endif
	else {
		_err_report("Unsupported color depth!");
		return NULL;
	}

	if(!sdl_initialized) {
		if (SDL_Init(SDL_INIT_VIDEO)) {
			_err_report("Unable to initialize SDL: %s", SDL_GetError());
			return NULL;
		}
	}
	sdl_initialized++;

	new = (_sdl_o *) malloc(sizeof(_sdl_o));
	if (new == NULL) {
		_err_report("Unable to allocate SDL object.");

		sdl_initialized--;
		if (!sdl_initialized)
			SDL_Quit();

		return NULL;
	}


	new->screen = SDL_SetVideoMode(width, height, depth,
		SDL_HWSURFACE | SDL_DOUBLEBUF);
	if (new->screen == NULL) {
		_err_report("Unable to open video surface: %s", SDL_GetError());

		sdl_initialized--;
		if (!sdl_initialized)
			SDL_Quit();
		free(new);

		return NULL;
	}

	new->image = SDL_CreateRGBSurface(SDL_HWSURFACE, width, height, depth,
		rmask, gmask, bmask, amask);

	if(new->image == NULL) {
		_err_report("Unable to create drawing surface: %s",
			SDL_GetError());

		SDL_FreeSurface(new->screen);

		sdl_initialized--;
		if (!sdl_initialized)
			SDL_Quit();
		free(new);
		
		return NULL;
	}

	new->depth = depth;

	if (depth == 32)
		new->pitch = new->image->pitch / 4;
	else if (depth == 16)
		new->pitch = new->image->pitch / 2;

	new->pixels.i32 = (uint32 *)new->image->pixels;

	return (sdl_o)new;
}

void sdl_close(sdl_o sdl)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;
	SDL_FreeSurface(_sdl->image);
	SDL_FreeSurface(_sdl->screen);

	sdl_initialized--;
	if (!sdl_initialized)
		SDL_Quit();
	free(_sdl);
}

uint32 sdl_color(sdl_o sdl, uint8 r, uint8 g, uint8 b)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;

	return (uint32)SDL_MapRGB(_sdl->image->format, r, g, b);
}

void sdl_putpixel(sdl_o sdl, uint x, uint y, uint32 color)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;

	if (_sdl->depth == 32)
		_sdl->pixels.i32[y*_sdl->pitch + x] = color;
	else if (_sdl->depth == 16)
		_sdl->pixels.i16[y*_sdl->pitch + x] = color;
}

void sdl_screenclr(sdl_o sdl, uint32 color)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;
	SDL_Rect rect;

	rect.x = 0;
	rect.y = 0;
	rect.w = _sdl->screen->w;
	rect.h = _sdl->screen->h;

	SDL_FillRect(_sdl->image, &rect, color);
}

void *sdl_getscreen(sdl_o sdl)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;

	return _sdl->pixels.i32;
}

void sdl_drawscreen(sdl_o sdl)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;
	SDL_Rect dest, src;

	dest.x = 0;
	dest.y = 0;
	dest.w = _sdl->screen->w;
	dest.h = _sdl->screen->h;

	src.x = 0;
	src.y = 0;
	src.w = _sdl->image->w;
	src.h = _sdl->image->h;


	SDL_BlitSurface(_sdl->image, &src, _sdl->screen, &dest);

	//SDL_UpdateRects(_sdl->screen, 1, &dest);

	SDL_Flip(_sdl->screen);
}

void sdl_grab(sdl_o sdl)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;

	SDL_WM_GrabInput(SDL_GRAB_ON);
}

void sdl_ungrab(sdl_o sdl)
{
	_sdl_o *_sdl = (_sdl_o *)sdl;

	SDL_WM_GrabInput(SDL_GRAB_OFF);
}
