#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>

#define VPCALC_SECTION	"LIB"
#define VPCALC_SUBSECT	"MEM"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/mem.h"

#include "../include/memory.h"

typedef struct _mem_o {
	void *library;

	memory_o memory;

	void (*del)(memory_o mem);

	size_t (*size)(memory_o mem);
	size_t (*bus_size)(memory_o mem);
	size_t (*addr_size)(memory_o mem);

	void (*putimage)(memory_o mem, void *data);
	void (*getimage)(memory_o mem, void *data);

	uint (*fetch)(memory_o mem, uint addr);
	uint (*read)(memory_o mem, uint addr);
	void (*write)(memory_o mem, uint addr, uint data);

	uint8 (*fetch_byte)(memory_o mem, uint addr);
	uint8 (*read_byte)(memory_o mem, uint addr);
	void (*write_byte)(memory_o mem, uint addr, uint8 data);

	void (*remap)(memory_o memory, uint page, uint addr, uint perm);
	int (*raw)(memory_o memory, void **rd_ptr, void **wr_ptr, void **ex_ptr);
} _mem_o;

mem_o mem_new(const char *libname)
{
	_mem_o *new;
	memory_o (*memory_new)(void);
	const char *error;

	new = (_mem_o *) malloc(sizeof(_mem_o));
	if (new == NULL) {
		_err_report("Unable to allocate memory structure.");

		return NULL;
	}

	new->library = dlopen(libname, RTLD_NOW);
	if (new->library == NULL) {
		error = dlerror();
		_err_report(error);

		free(new);
		return NULL;
	}

	memory_new = dlsym(new->library, "memory_new");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->del = dlsym(new->library, "memory_del");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->size = dlsym(new->library, "memory_size");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->bus_size = dlsym(new->library, "memory_bus_size");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->addr_size = dlsym(new->library, "memory_addr_size");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->fetch = dlsym(new->library, "memory_fetch");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->read = dlsym(new->library, "memory_read");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->write = dlsym(new->library, "memory_write");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->putimage = dlsym(new->library, "memory_putimage");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->getimage = dlsym(new->library, "memory_getimage");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->fetch_byte = dlsym(new->library, "memory_fetch_byte");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->read_byte = dlsym(new->library, "memory_read_byte");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->write_byte = dlsym(new->library, "memory_write_byte");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->remap = dlsym(new->library, "memory_remap");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->raw = dlsym(new->library, "memory_raw");
	error = dlerror();
	if (error != NULL) {
		_err_report(error);

		dlclose(new->library);
		free(new);
		return NULL;
	}

	new->memory = memory_new();
	if (new->memory == NULL) {
		_err_report("failed to make new memory unit.");

		dlclose(new->library);
		free(new);
		return NULL;
	}

	return (mem_o)new;
}

void mem_del(mem_o mem)
{
	_mem_o *_mem = (_mem_o *)mem;

	_mem->del(_mem->memory);

	dlclose(_mem->library);
	free(_mem);
}

size_t mem_size(mem_o mem)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->size(_mem->memory);
}

size_t mem_bus_size(mem_o mem)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->bus_size(_mem->memory);
}

size_t mem_addr_size(mem_o mem)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->addr_size(_mem->memory);
}

void mem_putimage(mem_o mem, void *data)
{
	_mem_o *_mem = (_mem_o *)mem;

	_mem->putimage(_mem->memory, data);
}

void mem_getimage(mem_o mem, void *data)
{
	_mem_o *_mem = (_mem_o *)mem;

	_mem->getimage(_mem->memory, data);
}

uint mem_fetch(mem_o mem, uint addr)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->fetch(_mem->memory, addr);
}

uint mem_read(mem_o mem, uint addr)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->read(_mem->memory, addr);
}

void mem_write(mem_o mem, uint addr, uint data)
{
	_mem_o *_mem = (_mem_o *)mem;

	_mem->write(_mem->memory, addr, data);
}

uint8 mem_fetch_byte(mem_o mem, uint addr)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->fetch_byte(_mem->memory, addr);
}

uint8 mem_read_byte(mem_o mem, uint addr)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->read_byte(_mem->memory, addr);
}

void mem_write_byte(mem_o mem, uint addr, uint8 data)
{
	_mem_o *_mem = (_mem_o *)mem;

	_mem->write_byte(_mem->memory, addr, data);
}

void mem_remap(mem_o mem, uint page, uint addr, uint perm)
{
	_mem_o *_mem = (_mem_o *)mem;

	_mem->remap(_mem->memory, page, addr, perm);
}
	
int mem_raw(mem_o mem, void **rd_ptr, void **wr_ptr, void **ex_ptr)
{
	_mem_o *_mem = (_mem_o *)mem;

	return _mem->raw(_mem->memory, rd_ptr, wr_ptr, ex_ptr);
}
